import sys
from copy import deepcopy

import numpy as np
import scipy.optimize
from matplotlib import pyplot as plt
from scipy.optimize import fsolve, minimize
from tqdm.auto import tqdm

sys.path.insert(0, '../')
from src.utils import get_colormaps, in_the_list, sort_eigs, make_orientation_consistent
import warnings

warnings.filterwarnings("ignore")
from sklearn.decomposition import PCA


class DynamicSystemAnalyzer():
    '''
    Generic class for analysis of the RNN dynamics: finding fixed points and plotting them in 2D and 3D
    '''

    def __init__(self, RNN_numpy):
        self.RNN = RNN_numpy
        self.rhs = self.RNN.rhs_noisless
        self.rhs_jac = self.RNN.rhs_jac
        self.fp_data = {}

        def objective(x, Input):
            return np.sum(self.rhs(x, Input) ** 2)

        def objective_grad(x, Input):
            return 2 * (self.rhs(x, Input).reshape(1, -1) @ self.rhs_jac(x, Input)).flatten()

        self.objective = objective
        self.objective_grad = objective_grad

    def get_fixed_points(self,
                         Input,
                         patience=100,
                         fun_tol=1e-12,
                         stop_length=100,
                         sigma_init_guess=10,
                         eig_cutoff=1e-10,
                         diff_cutoff=1e-7,
                         mode='exact'):
        '''
        calculates fixed points (stable, unstable and marginally stable)
        :param Input: np.array, input to the RNN at which the fixed points are calculated
        :param patience: the greater the patience parameter, the longer the fixed point are searched for. The search for
        the fixed points (FPs) terminates if patience is exceeded
        :param fun_tol: RHS norm tolerance parameter: the points with the values greate than fun_tol are discarded
        :param stop_length: the maximal number of fixed points. The search terminates if number of found FPs
        exceeds the stop_length parameter
        :param sigma_init_guess: the variance of the N-dimensional Gausssian distribution for generating an initial guess
        for a FP
        :param eig_cutoff: if the norm of the max eigenvalue of the Jacobian at a found FP is lesser than eig_cutoff,
        count this FP as "marginally stable"
        :param diff_cutoff: if the difference of the currently found FP with any of the previously found FPs
        (stored in the list) is lesser than diff_cutoff - discard this point as a duplicate
        :return: arrays of points: stable fixed points, unstable fixed points, marginally stable fixed points.
        with dimensions (num_points x N) in each array
        :param mode: 'exact' or 'approx'. 'exact' - computes exact fixed points, with scipy.optimize.fsolve method
        'approx' finds 'slow points' - points with small |RHS|^2, with fun_tol controlling the cut-off |RHS|^2.
        '''
        unstable_fps = [];
        stable_fps = [];
        marginally_stable_fps = [];
        all_points = []
        N = self.RNN.W_rec.shape[0]
        cntr = 0  # counter parameter, keeps track of how many times in a row an optimizer didn't find any new fp
        # proceed while (cntr <= patience) and unless one of the list start to overflow (because of a 2d attractor)
        while (cntr <= patience) and (len(all_points) < stop_length):
            x0 = sigma_init_guess * np.random.randn(N)
            # finding the roots of RHS of the RNN
            if mode == 'exact':
                x_root = fsolve(func=self.rhs, x0=x0, fprime=self.rhs_jac, args=(Input,))
            elif mode == "approx":
                res = scipy.optimize.minimize(fun=self.objective, x0=x0, args=(Input,), method='Powell')
                x_root = res.x
            else:
                raise ValueError(f"Mode {mode} is not implemented!")
            fun_val = self.objective(x_root, Input)
            cntr += 1
            if fun_val <= fun_tol:
                J = self.rhs_jac(x_root, Input)
                L = np.linalg.eigvals(J)
                L_0 = np.max(np.real(L))
                if not in_the_list(x_root, all_points, diff_cutoff=diff_cutoff):
                    cntr = 0
                    all_points.append(x_root)
                    if (np.abs(L_0) <= eig_cutoff):  # marginally stable fixed point (belongs to 1D attractor)
                        marginally_stable_fps.append(x_root)
                    else:
                        stable_fps.append(x_root) if (L_0 < -eig_cutoff) else unstable_fps.append(x_root)
        # Saving the data in the internal dictionary accessible by the input vector turned into string:
        input_as_key = str(Input.tolist())
        self.fp_data[input_as_key] = {}
        point_types = ["stable_fps", "unstable_fps", "marginally_stable_fps"]
        for i, type in enumerate(point_types):
            if len(eval(type)) != 0:
                self.fp_data[input_as_key][type] = np.vstack(eval(point_types[i]))
        return None

    def plot_fixed_points(self, projection='2D', P=None):
        '''
        a function plots all the pre-calculated fixed-points
        If projection matrix P is None it performs PCA on the points
        and then plots these points projected on the first PCs
        If P is supplied, projects the points using the matrix P
        if not, performs PCA on the recovered points and uses first PCs for projecting.

        :param projection: to plot the FPs either on a plane (2D) or in 3D
        :param P: projection matrix N x n_dim size
        :return: a figure of fixed points on the first
        '''
        n_dim = 2 if projection == '2D' else 3
        inputs_as_key = list(self.fp_data.keys())
        if len(inputs_as_key) == 0:
            raise ValueError("To plot fixed points, one has to calculate them using 'get_fixed_point function' first!")
        all_points = []
        for input_as_key in inputs_as_key:
            types = list(self.fp_data[input_as_key].keys())
            points_per_input = np.vstack([self.fp_data[input_as_key][type] for type in types])
            all_points.append(points_per_input)
        all_points = np.vstack(all_points)
        if all_points.shape[0] < n_dim:
            raise ValueError("The number of found fixed points is lesser than n_dim of projection!")

        if P is None:
            pca = PCA(n_components=n_dim)
            pca.fit(all_points)
            P = np.array(pca.components_).T
        # projecting fixed points onto n_dim-subspace
        data_to_plot = {}
        for input_as_key in inputs_as_key:
            data_to_plot[input_as_key] = {}
            types = list(self.fp_data[input_as_key].keys())
            for type in types:
                data_to_plot[input_as_key][type] = self.fp_data[input_as_key][type] @ P

        color_sets = [["blue", "red", "black"], ["blueviolet", "tomato", "darkslategray"],
                      ["green", "darkorange", "midnightblue"]]

        # Plotting the fixed points
        if n_dim == 2:
            fig = plt.figure(figsize=(7, 7))
            fig.suptitle(r"Fixed points projected on 2D PCA plane", fontsize=16)
            for j, input_as_key in enumerate(inputs_as_key):
                markers = ["o", "x", "o"];
                colors = color_sets[j]
                types = list(data_to_plot[input_as_key].keys())
                for t, type in enumerate(types):
                    plt.scatter(data_to_plot[input_as_key][type][:, 0],
                                data_to_plot[input_as_key][type][:, 1],
                                marker=markers[t], s=100, color=colors[t],
                                edgecolors='k')
            plt.xlabel("PC 1", fontsize=16)
            plt.ylabel("PC 2", fontsize=16)
            plt.grid(True)
        elif n_dim == 3:
            fig = plt.figure(figsize=(7, 7))
            ax = fig.add_subplot(projection='3d')
            # make the panes transparent
            ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.set_xlabel("PC 1", fontsize=20)
            ax.set_ylabel("PC 2", fontsize=20)
            ax.set_zlabel("PC 3", fontsize=20)
            fig.suptitle(r"Fixed points projected on 3D PCA subspace", fontsize=16)
            for j, input_as_key in enumerate(inputs_as_key):
                markers = ["o", "x", "o"];
                colors = color_sets[j]
                types = list(data_to_plot[input_as_key].keys())
                for t, type in enumerate(types):
                    ax.scatter(data_to_plot[input_as_key][type][:, 0],
                               data_to_plot[input_as_key][type][:, 1],
                               data_to_plot[input_as_key][type][:, 2],
                               marker=markers[t], s=100, color=colors[t],
                               edgecolors='k')
            plt.grid(True)
        return fig

    def compute_point_analytics(self, point, Input):
        '''
        :param point: a numpy vector, specifying a point of the state-space of RNN
        :param Input: a numpy vector, specifying the input to the RNN
        :return: fun_val - value of the |RHS|^2, J - Jacobian, E - eigenvalues sorted according their real part,
        the first left eigenvector l and the first right eigenvector r
        '''
        self.RNN.y = deepcopy(point)
        fun_val = self.objective(point, Input)
        J = self.rhs_jac(point, Input)
        E, R = np.linalg.eig(J)
        E, R = sort_eigs(E, R)
        L = np.linalg.inv(R)
        l = np.real(L[0, :])
        r = np.real(R[:, 0])
        return fun_val, J, E, l, r


class DynamicSystemAnalyzerCDDM(DynamicSystemAnalyzer):
    '''
    Class which is inрerited from the DynamicSystemAnalyzer base class,
    dedicated to processing of the RNNs trained on CDDM task
    '''

    def __init__(self, RNN):
        DynamicSystemAnalyzer.__init__(self, RNN)
        self.choice_axis = self.RNN.W_out.flatten() if self.RNN.W_out.shape[0] == 1 \
            else (self.RNN.W_out[0, :] - self.RNN.W_out[1, :])
        self.context_axis = self.RNN.W_inp[:, 0] - self.RNN.W_inp[:, 1]
        self.sensory_axis = np.sum([self.RNN.W_inp[:, i] for i in [2, 3, 4, 5]])

        self.LA_data = {}
        self.LA_data["motion"] = {}
        self.LA_data["color"] = {}

    def get_LineAttractor_endpoints(self, context, nudge=0.05, T_steps=1000, relax_steps=10):
        '''
        :param context: "motion" or "color"
        :param nudge: an additional input, creating a bias either to the right choice, or to the left choice
        :param T_steps: run the RNN with the applied input for the T_steps duration
        :param relax_steps: after T_steps with input, turn off any input and let the RNN relax to 'slow points'
        :return: left-most and right-most points of a presupposed line attractor
        '''
        ctxt_ind = 0 if context == 'motion' else 1
        sensory_inds = [2, 3] if context == 'motion' else [4, 5]
        default_input = np.array([0, 0, 0.5, 0.5, 0.5, 0.5])
        default_input[ctxt_ind] = 1.0
        input_right_decision = deepcopy(default_input)
        input_left_decision = deepcopy(default_input)
        input_right_decision[sensory_inds] += np.array([nudge, -nudge])
        input_left_decision[sensory_inds] -= np.array([nudge, -nudge])
        points = []
        # find left and right points
        for inp in [input_left_decision, input_right_decision]:
            self.RNN.y_init = deepcopy(self.RNN.y_init)
            self.RNN.y = deepcopy(self.RNN.y_init)
            self.RNN.run(input_timeseries=np.repeat(inp[:, np.newaxis], axis=1, repeats=T_steps))
            self.RNN.run(input_timeseries=np.repeat(inp[:, np.newaxis], axis=1, repeats=relax_steps))
            points.append(deepcopy(self.RNN.y))
        return points[0], points[1]

    def calc_LineAttractor_analytics(self,
                                     N_points=31,
                                     obj_max_iter=100,
                                     nudge=0.05,
                                     T_steps=1000,
                                     relax_steps=10):
        '''
        :param N_points: number of points on each line attractor
        :param obj_max_iter:  maximum iteration in the |RHS|^2 minimization process
        the rest of the parameters are the same as in 'get_LineAttractor_endpoints', 'get_fixed_points'
        :return: a dictionary with "color" and "motion" contexts, each containing sub-dictionary with:
        'slow points', |RHS|^2 value, Jacobian, eigenvalues, the principal left and right eigenvectors over these points
        '''
        default_input = np.array([0, 0, 0.5, 0.5, 0.5, 0.5])
        for context in ["motion", "color"]:
            ctxt_ind = 0 if context == 'motion' else 1
            Input = deepcopy(default_input)
            Input[ctxt_ind] = 1
            # get the end points of the line attractor
            left_point, right_point = self.get_LineAttractor_endpoints(context,
                                                                       nudge=nudge,
                                                                       T_steps=T_steps,
                                                                       relax_steps=relax_steps)
            # define the starting direction for search
            increment = (1 / (N_points - 1)) * (right_point - left_point)
            direction = deepcopy(increment)
            direction *= np.sign(np.dot(self.choice_axis, direction))
            slow_points = [];
            fun_vals = [];
            eigs = [];
            jacs = [];
            selection_vects = [];
            principle_eigenvects = []
            x_init = deepcopy(left_point)
            print(f"Analyzing points on a line attractor in {context} context...")
            for i in tqdm(range(N_points)):
                # minimize ||RHS(x)|| such that the x stays within a space orthogonal to the line attractor
                res = minimize(self.objective, x0=x_init, args=(Input,), method='SLSQP',
                               jac=self.objective_grad, options={'disp': False, 'maxiter': obj_max_iter},
                               constraints={'type': 'eq', 'fun': lambda x: np.dot(x - x_init, increment)})
                x_root = deepcopy(res.x)
                x_init = x_root + increment
                slow_pt = deepcopy(x_root)

                # compute analytics at the slow point:
                fun_val, J, E, l, r = self.compute_point_analytics(slow_pt, Input)
                k = np.sign(np.dot(r, direction))
                l *= k;
                r *= k

                slow_points.append(deepcopy(slow_pt))
                fun_vals.append(fun_val)
                jacs.append(deepcopy(J))
                eigs.append(deepcopy(E))
                selection_vects.append(deepcopy(l))
                principle_eigenvects.append(deepcopy(r))
            selection_vects = make_orientation_consistent(selection_vects)
            principle_eigenvects = make_orientation_consistent(principle_eigenvects)

            self.LA_data[context]["slow_points"] = (np.array(slow_points))
            self.LA_data[context]["fun_val"] = (np.array(fun_vals))
            self.LA_data[context]["jac"] = (np.array(jacs))
            self.LA_data[context]["eigs"] = (np.array(eigs))
            self.LA_data[context]["l"] = (np.array(selection_vects))
            self.LA_data[context]["r"] = (np.array(principle_eigenvects))
        return None

    def plot_LineAttractor_3D(self,
                              nudge=0.05,
                              steps_stim_on=500,
                              steps_context_only_on=250):
        '''
        Plots
        Figure with projected line attractors within each context on a (sensory, context, choice)-subspace
        also plots 6 trajectories: three in each context.
        light blue - color context, orange - motion context,
        solid lines - trajectories with the decision to the right, dashed lines - to the left.
        two magenta trajectores - trials within each context with 0 coherence
        :return:
        '''
        nDim = 3
        # projection matrix
        P_matrix = np.zeros((self.RNN.N, 3))
        P_matrix[:, 0] = self.choice_axis
        P_matrix[:, 1] = self.context_axis
        P_matrix[:, 2] = self.sensory_axis
        if not ("slow_points" in (list(self.LA_data["motion"].keys()))):
            raise ValueError("Need to run `calc_LineAttractor_analytics' first!")
        else:
            LA_data_dict = deepcopy(self.LA_data)

        trajectories = dict()
        trajectories["motion"] = {}
        trajectories["color"] = {}
        for ctxt in ["motion", "color"]:
            trajectories[ctxt] = {}
            for stim_status in ["relevant", "irrelevant"]:
                trajectories[ctxt][stim_status] = {}
                for period in ["context_only_on", "stim_on", "stim_off"]:
                    trajectories[ctxt][stim_status][period] = {}

        colors, cmp = get_colormaps()
        red, blue, bluish, green, orange, lblue, violet = colors
        colors_trajectories = dict()
        colors_trajectories["motion"] = dict()
        colors_trajectories["color"] = dict()
        colors_trajectories["motion"]["relevant"] = colors[5]
        colors_trajectories["motion"]["irrelevant"] = colors[3]
        colors_trajectories["color"]["relevant"] = colors[1]
        colors_trajectories["color"]["irrelevant"] = colors[3]

        for ctxt in ["motion", "color"]:
            val = 1 if ctxt == 'motion' else 0
            for stim_status in ["relevant", "irrelevant"]:
                self.RNN.clear_history()
                rel_inds = [2, 3] if ctxt == 'motion' else [4, 5]
                irrel_inds = [4, 5] if ctxt == 'motion' else [2, 3]
                nudge_inds = rel_inds if stim_status == 'relevant' else irrel_inds

                x0 = 0.00 * np.random.randn(self.RNN.N)
                input = np.array([val, 1 - val, 0.0, 0.0, 0.0, 0.0])
                input_timeseries = np.repeat(input[:, np.newaxis], axis=1, repeats=steps_context_only_on)
                self.RNN.y = deepcopy(x0)
                self.RNN.run(input_timeseries=input_timeseries, save_history=True)
                x_trajectory_context_only_on = self.RNN.get_history()
                trajectories[ctxt][stim_status]["context_only_on"] = x_trajectory_context_only_on
                self.RNN.clear_history()

                x0 = deepcopy(x_trajectory_context_only_on[-1, :])
                for direction in ['left', 'right', 'center']:
                    input = deepcopy(np.array([val, 1 - val, 0.5, 0.5, 0.5, 0.5]))
                    if direction == 'left':
                        input[nudge_inds] -= np.array([nudge, -nudge])
                    elif direction == 'right':
                        input[nudge_inds] += np.array([nudge, -nudge])
                    input_timeseries = np.repeat(input[:, np.newaxis], axis=1, repeats=steps_stim_on)
                    self.RNN.y = deepcopy(x0)
                    self.RNN.run(input_timeseries=input_timeseries, save_history=True)
                    trajectory = self.RNN.get_history()
                    trajectories[ctxt][stim_status]["stim_on"][direction] = deepcopy(trajectory)
                    self.RNN.clear_history()

        colors_trajectories = dict()
        colors_trajectories["motion"] = dict()
        colors_trajectories["color"] = dict()
        colors_trajectories["motion"]["relevant"] = orange
        colors_trajectories["motion"]["irrelevant"] = orange
        colors_trajectories["color"]["relevant"] = lblue
        colors_trajectories["color"]["irrelevant"] = lblue
        colors_LA = dict()
        colors_LA["motion"] = bluish
        colors_LA["color"] = green

        fig_3D = plt.figure(figsize=(7, 7))
        ax = fig_3D.add_subplot(projection='3d')
        # make the panes transparent
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.set_xlabel("Choice", fontsize=20)
        ax.set_ylabel("Context", fontsize=20)
        ax.set_zlabel("Sensory", fontsize=20)

        # initial point trajectory
        ax.scatter([0, 0], [0, 0], [0, 0], color='r', marker='o', s=10, alpha=0.9)

        for ctxt in ["motion", "color"]:
            slow_points_projected = LA_data_dict[ctxt]["slow_points"] @ P_matrix
            ax.scatter(*(slow_points_projected[:, k] for k in range(nDim)), color=colors_LA[ctxt], marker='o', s=20,
                       alpha=0.5)
            ax.plot(*(slow_points_projected[:, k] for k in range(nDim)), color=colors_LA[ctxt])
            for stim_status in ["relevant"]:
                clr = colors_trajectories[ctxt][stim_status]
                trajectory_projected = trajectories[ctxt][stim_status]["context_only_on"] @ P_matrix
                ax.plot(*(trajectory_projected[:, t] for t in range(nDim)),
                        linestyle='-', linewidth=2, color=clr, alpha=0.8)
                linestyles = ['-', ':', '-']
                colors = [clr, clr, 'm']
                for k, key in enumerate(["right", "left", "center"]):
                    trajectory_projected = trajectories[ctxt][stim_status]["stim_on"][key] @ P_matrix
                    ax.plot(*(trajectory_projected[:, t] for t in range(nDim)),
                            linestyle=linestyles[k], linewidth=3, color=colors[k], alpha=0.8)

        fig_3D.subplots_adjust()
        ax.view_init(12, 228)
        fig_3D.subplots_adjust()
        plt.tight_layout()
        return fig_3D

    def plot_RHS_over_LA(self):
        if not ("slow_points" in (list(self.LA_data["motion"].keys()))):
            raise ValueError("Need to run `calc_LineAttractor_analytics' first!")
        else:
            LA_data_dict = deepcopy(self.LA_data)
        colors, cmp = get_colormaps()
        red, blue, bluish, green, orange, lblue, violet = colors
        fig_RHS = plt.figure(figsize=(12, 3))
        plt.suptitle(r"$\||RHS(x)\||^2$", fontsize=16)
        plt.axhline(0, color="gray", linewidth=2, alpha=0.2)
        x = np.linspace(0, 1, LA_data_dict['motion']["slow_points"].shape[0])
        plt.plot(x, np.array(LA_data_dict["motion"]["fun_val"]), color=bluish, linewidth=3, linestyle='-',
                 label="motion")
        plt.plot(x, np.array(LA_data_dict["color"]["fun_val"]), color=green, linewidth=3, linestyle='-', label="color")
        plt.legend(fontsize=14)
        plt.xlabel("distance along the LA", fontsize=16)
        plt.ylabel(r"$\||RHS(x)\||$", fontsize=16)
        plt.grid(True)
        return fig_RHS



import json
import os
import sys

import numpy as np

sys.path.insert(0, '../')
sys.path.insert(0, '../../')
from src.utils import get_project_root
from datetime import date

date = ''.join((list(str(date.today()).split("-"))[::-1]))

# RNN specific
N = 100
activation_name = 'relu'
constrained = True
seed = None
sigma_inp = 0.03
sigma_rec = 0.03
dt = 1
tau = 10
sr = 1.2
connectivity_density_rec = 1.0

# task specific
task_name = 'CDDM'
n_inputs = 6
n_outputs = 2
T = 750
n_steps = int(T / dt)
max_coherence = 0.8
coherence_lvls = 5

mask = np.concatenate([np.arange(int(n_steps // 3)), int(2 * n_steps // 3) + np.arange(int(n_steps // 3))]).tolist()
task_params = {"cue_on": 0, "cue_off": n_steps,
               "stim_on": int(n_steps // 3), "stim_off": n_steps,
               "dec_on": int(2 * n_steps // 3), "dec_off": n_steps,
               "n_steps": n_steps, "n_inputs": n_inputs, "n_outputs": n_outputs}
tmp = max_coherence * np.logspace(-(coherence_lvls - 1), 0, coherence_lvls, base=2)
coherences = np.concatenate([-np.array(tmp[::-1]), np.array([0]), np.array(tmp)]).tolist()
task_params["coherences"] = coherences
task_params["seed"] = seed

# training specific
max_iter = 1000
tol = 1e-10
lr = 0.02
weight_decay = 5e-6
lambda_orth = 0.3
lambda_r = 0.3
same_batch = True
shuffle = False

data_folder = os.path.abspath(os.path.join(get_project_root(), "data", "trained_RNNs", f"{task_name}"))
tag = f'{task_name}_{activation_name}'

config_dict = {}
config_dict["N"] = N
config_dict["seed"] = seed
config_dict["activation"] = activation_name
config_dict["sigma_inp"] = sigma_inp
config_dict["sigma_rec"] = sigma_rec
config_dict["num_inputs"] = n_inputs
config_dict["num_outputs"] = n_outputs
config_dict["constrained"] = constrained
config_dict["dt"] = dt
config_dict["tau"] = tau
config_dict["sr"] = sr
config_dict["connectivity_density_rec"] = connectivity_density_rec
config_dict["max_iter"] = max_iter
config_dict["n_steps"] = n_steps
config_dict["task_params"] = task_params
config_dict["mask"] = mask
config_dict["tol"] = tol
config_dict["lr"] = lr
config_dict["same_batch"] = same_batch
config_dict["weight_decay"] = weight_decay
config_dict["lambda_orth"] = lambda_orth
config_dict["lambda_r"] = lambda_r
config_dict["data_folder"] = data_folder
config_dict["tag"] = tag
config_dict["shuffle"] = shuffle

json_obj = json.dumps(config_dict, indent=4)
outfile = open(os.path.join(get_project_root(), "data", "configs", f"train_config_{tag}.json"), mode="w")
outfile.write(json_obj)



import json
import os
import sys
from datetime import date

import numpy as np

sys.path.insert(0, '../')
sys.path.insert(0, '../../')
from src.utils import get_project_root

date = ''.join((list(str(date.today()).split("-"))[::-1]))

# RNN specific
N = 50
activation_name = 'tanh'
constrained = False
seed = None
sigma_inp = 0.015
sigma_rec = 0.015
dt = 1
tau = 10
sr = 1.2
connectivity_density_rec = 1.0

# task specific
task_name = 'MemoryAntiNumber'
n_inputs = 2
n_outputs = 1
T = 320
n_steps = int(T / dt)
task_params = dict()
task_params["stim_on_range"] = [0, 8 * n_steps // 16 - 20]
task_params["stim_duration"] = 10
task_params["recall_on"] = 8 * n_steps // 16
task_params["recall_off"] = n_steps
task_params["seed"] = seed
# mask = np.concatenate([np.arange(n_steps)]).tolist() # using the whole trial
mask = np.concatenate([np.arange(task_params["recall_on"]),
                       9 * n_steps // 16 + np.arange(7 * n_steps // 16)]).tolist()  # using the whole trial

# training specific
max_iter = 800
tol = 1e-10
lr = 0.005
weight_decay = 5e-6
lambda_orth = 0.3
lambda_r = 0.1
same_batch = False  # generate new batch in each train loop
shuffle = False

data_folder = os.path.abspath(os.path.join(get_project_root(), "data", "trained_RNNs", f"{task_name}"))
tag = f'{task_name}_{activation_name}'

config_dict = {}
config_dict["N"] = N
config_dict["seed"] = seed
config_dict["activation"] = activation_name
config_dict["sigma_inp"] = sigma_inp
config_dict["sigma_rec"] = sigma_rec
config_dict["num_inputs"] = n_inputs
config_dict["num_outputs"] = n_outputs
config_dict["constrained"] = constrained
config_dict["dt"] = dt
config_dict["tau"] = tau
config_dict["sr"] = sr
config_dict["connectivity_density_rec"] = connectivity_density_rec
config_dict["max_iter"] = max_iter
config_dict["n_steps"] = n_steps
config_dict["task_params"] = task_params
config_dict["mask"] = mask
config_dict["tol"] = tol
config_dict["lr"] = lr
config_dict["same_batch"] = same_batch
config_dict["weight_decay"] = weight_decay
config_dict["lambda_orth"] = lambda_orth
config_dict["lambda_r"] = lambda_r
config_dict["data_folder"] = data_folder
config_dict["tag"] = tag

json_obj = json.dumps(config_dict, indent=4)
outfile = open(os.path.join(get_project_root(), "data", "configs", f"train_config_{tag}.json"), mode="w")
outfile.write(json_obj)



import sys

sys.path.insert(0, "../")
import numpy as np
from copy import deepcopy
import numdifftools as nd

'''
lightweight numpy implementation of RNN for validation and quick testing and plotting
'''


def ReLU(x):
    return np.maximum(x, 0)


class RNN_numpy():
    def __init__(self, N, dt, tau, W_inp, W_rec, W_out, bias_rec=None, activation=ReLU, y_init=None):
        self.N = N
        self.W_inp = W_inp
        self.W_rec = W_rec
        self.W_out = W_out
        if bias_rec is None:
            self.bias_rec = np.zeros(self.N)
        else:
            self.bias_rec = bias_rec
        self.dt = dt
        self.tau = tau
        self.alpha = self.dt / self.tau
        if not (y_init is None):
            self.y_init = y_init
        else:
            self.y_init = np.zeros(self.N)
        self.y = deepcopy(self.y_init)
        self.y_history = []
        self.activation = activation

    def rhs(self, y, input, sigma_rec=None, sigma_inp=None, generator_numpy=None):
        if (generator_numpy is None): generator_numpy = np.random.default_rng(np.random.randint(10000))
        if len(y.shape) == 2:
            # Check that the batch_size (last dimension) is the same as the Input's last dimension
            if y.shape[-1] != input.shape[-1]:
                raise ValueError(
                    f"The last dimension of the RNN state and the Input (representing batch size) should be equal!" +
                    f" {x.shape[-1]} != {input.shape[-1]}")
            batch_size = y.shape[-1]
            bias_rec = np.repeat(self.bias_rec[:, np.newaxis], repeats=batch_size, axis=-1)
        else:
            bias_rec = self.bias_rec

        if ((sigma_rec is None) and (sigma_inp is None)) or ((sigma_rec == 0) and (sigma_inp == 0)):
            return -y + self.activation(self.W_rec @ y + self.W_inp @ input + self.bias_rec)
        else:
            rec_noise_term = np.sqrt((2 / self.alpha) * sigma_rec ** 2) * generator_numpy.standard_normal(y.shape) \
                if (not (sigma_rec is None)) else np.zeros(x.shape)
            inp_noise_term = np.sqrt((2 / self.alpha) * sigma_inp ** 2) * generator_numpy.standard_normal(input.shape) \
                if (not (sigma_inp is None)) else np.zeros(input.shape)
            return -y + self.activation(
                self.W_rec @ y + self.W_inp @ (input + inp_noise_term) + bias_rec + rec_noise_term)

    def rhs_noisless(self, y, input):
        '''
        Bare version of RHS for efficient fixed point analysis
        supposed to work only with one point at the state-space at the time (no batches!)
        '''
        return -y + self.activation(self.W_rec @ y + self.W_inp @ input + self.bias_rec)

    def rhs_jac(self, y, input):
        # efficient calculation of Jacobian using a finite difference (lesser hustle than with autograd!)
        return nd.Jacobian(self.rhs_noisless)(y, input)

    def step(self, input, sigma_rec=None, sigma_inp=None, generator_numpy=None):
        self.y += (self.dt / self.tau) * self.rhs(self.y, input, sigma_rec, sigma_inp, generator_numpy)

    def run(self, input_timeseries, save_history=False, sigma_rec=None, sigma_inp=None, generator_numpy=None):
        '''
        :param Inputs: an array, has to be iether (n_inputs x n_steps) dimensions or (n_inputs x n_steps x batch_batch_size)
        :param save_history: bool, whether to save the resulting trajectory
        :param sigma_rec: noise parameter in the recurrent dynamics
        :param sigma_inp: noise parameter in the input channel
        :param generator_numpy: numpy random number generator, for reproducibility
        :return: None
        '''
        num_steps = input_timeseries.shape[1]  # second dimension
        if len(input_timeseries.shape) == 3:
            batch_size = input_timeseries.shape[-1]  # last dimension
            # if the state is a 1D vector, repeat it batch_size number of times to match with Input dimension
            if len(self.y.shape) == 1:
                self.y = np.repeat(deepcopy(self.y)[:, np.newaxis], axis=1, repeats=batch_size)
        for i in range(num_steps):
            if save_history == True:
                self.y_history.append(deepcopy(self.y))
            self.step(input_timeseries[:, i, ...], sigma_rec=sigma_rec, sigma_inp=sigma_inp,
                      generator_numpy=generator_numpy)
        return None

    def get_history(self):
        return np.array(self.y_history)

    def clear_history(self):
        self.y_history = []

    def reset_state(self):
        self.y = np.zeros(self.N)

    def get_output(self):
        y_history = np.stack(self.y_history, axis=0)
        output = np.swapaxes((self.W_out @ y_history), 0, 1)
        return output

    def run_multiple_trajectories(self, input_timeseries, sigma_rec, sigma_inp, generator_numpy=None):
        if type(input_timeseries) == list:
            input_timeseries = np.array(input_timeseries)
        # Check that inputs have 3 dimensions: n_inputs x n_steps x batch_size
        if len(input_timeseries.shape) != 3:
            raise ValueError("Inputs dimension have to be n_inputs x n_steps x batch_size!")
        self.clear_history()
        batch_size = input_timeseries.shape[-1]
        self.y = np.repeat(deepcopy(self.y_init)[:, np.newaxis], axis=-1, repeats=batch_size)
        self.run(input_timeseries=input_timeseries,
                 sigma_rec=sigma_rec,
                 sigma_inp=sigma_inp,
                 save_history=True,
                 generator_numpy=generator_numpy)
        trajectories = np.stack(self.y_history, axis=1)  # N x n_steps x batch_size
        outputs = self.get_output()  # n_outputs x n_steps x batch_size
        return trajectories, outputs


if __name__ == '__main__':
    N = 100
    x = np.random.randn(N)
    W_rec = np.random.randn(N, N)
    W_inp = np.random.randn(N, 6)
    W_out = np.random.randn(2, N)
    bias_rec = np.random.randn(N)

    # Input = np.ones(6)
    dt = 0.1
    tau = 10
    batch_size = 1
    input = np.ones((6, batch_size))
    activation_fun = np.tanh
    rnn = RNN_numpy(N=N, W_rec=W_rec, W_inp=W_inp, W_out=W_out, dt=dt, tau=tau, activation=activation_fun)
    rnn.y = np.random.randn(N)
    print(rnn.rhs_noisless(rnn.y, input=input).shape)
    print((rnn.rhs_jac(rnn.y, input=input)))
    # generate multidimensional inputs
    # input_timeseries = 0.1 * np.ones((6, 301, 32))
    # rnn.run(input_timeseries=input_timeseries, sigma_rec=0.03, sigma_inp=0.03)
    # trajectories, outputs = rnn.run_multiple_trajectories(input_timeseries=input_timeseries,
    #                                                       sigma_rec=0.03,
    #                                                       sigma_inp=0.03)
    # print(trajectories.shape)
    # print(outputs.shape)



import hashlib
import json
import os
import pickle
import sys
from copy import deepcopy

import numpy as np

sys.path.insert(0, "../")
sys.path.insert(0, "../../")
from src.datajoint.datajoint_config import *


def jsonify(dct):
    dct_jsonified = {}
    for key in list(dct.keys()):
        if type(dct[key]) == type({}):
            dct_jsonified[key] = jsonify(dct[key])
        elif type(dct[key]) == np.ndarray:
            dct_jsonified[key] = dct[key].tolist()
        else:
            dct_jsonified[key] = dct[key]
    return dct_jsonified


def get_hash(dictionary):
    dictionary_json = json.dumps(dictionary, sort_keys=True).encode("utf-8")
    h = hashlib.md5(dictionary_json).hexdigest()
    return h


folder = os.path.join("../", 'data', 'trained_RNNs', 'CDDM')
subfolders = os.listdir(folder)
for subfolder in subfolders:
    if not subfolder.startswith('.'):
        for file in list(os.listdir(os.path.join(folder, subfolder))):
            if 'params' in file:
                score = file.split('_')[-1].split(".pkl")[0]
                taskname = file.split('_')[1]

        config_dict = json.load(open(os.path.join(folder, subfolder, f'{score}_config.json'), 'rb'))
        net_params_file = os.path.join(folder, subfolder, f'params_CDDM_{score}.pkl')
        net_params = pickle.load(open(net_params_file, 'rb'))

        LA_data_file = os.path.join(folder, subfolder, f'{score}_LA_data.pkl')
        LA_data_file_exists = os.path.exists(LA_data_file)
        LA_data = pickle.load(open(LA_data_file, 'rb')) if LA_data_file_exists else None

        fp_data_file = os.path.join(folder, subfolder, f'{score}_fp_data.pkl')
        fp_data_file_exists = os.path.exists(fp_data_file)
        fp_data = pickle.load(open(fp_data_file, 'rb')) if fp_data_file_exists else None

        psycho_data_file = os.path.join(folder, subfolder, f'{score}_psycho_data.pkl')
        psycho_data_file_exists = os.path.exists(fp_data_file)
        psycho_data = pickle.load(open(psycho_data_file, 'rb')) if psycho_data_file_exists else None

        #         if psycho_data is None: print('found no psycho_data file!')
        #         if fp_data is None: print('found no fp_data file!')
        #         if LA_data is None: print('found no LA_data file!')

        if (psycho_data is None) or (fp_data is None) or (LA_data is None):
            pass  # skip this entry
        else:
            rnn_dj = RNNDJ()
            task_dj = TaskDJ()
            trainer_dj = TrainerDJ()
            if psycho_data_file_exists or fp_data_file_exists or LA_data_file_exists:
                cddm_analysis_dj = CDDMRNNAnalysisDJ()
            else:
                cddm_analysis_dj = None
            config_dict["task_params"]["seed"] = None  # non randomized task!
            task_dj_dict = {"task_name": taskname,
                            "n_steps": config_dict["n_steps"],
                            "n_inputs": config_dict["num_inputs"],
                            "n_outputs": config_dict["num_outputs"],
                            "task_params": config_dict["task_params"],
                            "mask": config_dict["mask"]}
            task_hash = get_hash(jsonify(task_dj_dict))
            task_dj_dict["task_hash"] = task_hash

            trainer_dj_dict = {"max_iter": config_dict["max_iter"],
                               "tol": config_dict["tol"],
                               "lr": config_dict["lr"],
                               "lambda_orth": config_dict["lambda_orth"],
                               "lambda_r": config_dict["lambda_r"],
                               "same_batch": config_dict["same_batch"],
                               "shuffle": False}
            trainer_hash = get_hash(jsonify(trainer_dj_dict))
            trainer_dj_dict["trainer_hash"] = trainer_hash

            rnn_dj_dict = {"task_name": taskname,
                           "mse_score": score,
                           "n": config_dict["N"],
                           "activation_name": config_dict["activation"],
                           "constrained": config_dict["constrained"],
                           "dt": config_dict["dt"],
                           "tau": config_dict["tau"],
                           "sr": config_dict["sr"],
                           "connectivity_density_rec": config_dict["connectivity_density_rec"],
                           "sigma_rec": config_dict["sigma_rec"],
                           "sigma_inp": config_dict["sigma_inp"],
                           "task_hash": task_hash,
                           "trainer_hash": trainer_hash}
            rnn_dj_dict["w_inp"] = net_params["W_inp"]
            rnn_dj_dict["w_rec"] = net_params["W_rec"]
            rnn_dj_dict["w_out"] = net_params["W_out"]
            rnn_dj_dict["b_rec"] = 0 if net_params["bias_rec"] is None else net_params["bias_rec"]
            rnn_hash = get_hash(jsonify(rnn_dj_dict))
            rnn_dj_dict["rnn_hash"] = rnn_hash
            cddm_analysis_dj_dict = {"task_hash": task_hash,
                                     "trainer_hash": trainer_hash,
                                     "rnn_hash": rnn_hash,
                                     "task_name": taskname,
                                     "mse_score": score,
                                     "psycho_data": deepcopy(psycho_data),
                                     "fp_data": deepcopy(fp_data),
                                     "la_data": deepcopy(LA_data)}

            task_dj.insert1(task_dj_dict, skip_duplicates=True)
            trainer_dj.insert1(trainer_dj_dict, skip_duplicates=True)
            rnn_dj.insert1(rnn_dj_dict, skip_duplicates=True)
            if not (cddm_analysis_dj is None):
                cddm_analysis_dj.insert1(cddm_analysis_dj_dict, skip_duplicates=True)

            print(f"successfully created new entry!")



from copy import deepcopy

import numpy as np
from matplotlib import pyplot as plt


class PerformanceAnalyzer():
    '''
    Generic class for analysis of the RNN performance on the given task
    '''

    def __init__(self, rnn_numpy, task=None):
        self.RNN = rnn_numpy
        self.Task = task

    def get_validation_score(self, scoring_function,
                             input_batch, target_batch, mask,
                             sigma_rec=0.03, sigma_inp=0.03,
                             rng_numpy=None):
        n_inputs = input_batch.shape[0]
        n_steps = input_batch.shape[1]
        batch_size = input_batch.shape[2]
        self.RNN.clear_history()
        # self.RNN.y = np.repeat(deepcopy(self.RNN.y)[:, np.newaxis], axis=-1, repeats=batch_size)
        trajectories, output_prediction = self.RNN.run_multiple_trajectories(input_timeseries=input_batch,
                                                                             sigma_rec=sigma_rec,
                                                                             sigma_inp=sigma_inp,
                                                                             generator_numpy=rng_numpy)
        avg_score = np.mean(
            [scoring_function(output_prediction[:, mask, i], target_batch[:, mask, i]) for i in range(batch_size)])
        return avg_score

    def plot_trials(self, input_batch, target_batch, mask, sigma_rec=0.03, sigma_inp=0.03, labels=None, rng_numpy=None):
        n_inputs = input_batch.shape[0]
        n_steps = input_batch.shape[1]
        batch_size = input_batch.shape[2]
        fig_output, axes = plt.subplots(batch_size, 1, figsize=(7, 8))
        self.RNN.clear_history()
        self.RNN.y = deepcopy(self.RNN.y_init)
        self.RNN.run_multiple_trajectories(input_timeseries=input_batch,
                                           sigma_rec=sigma_rec,
                                           sigma_inp=sigma_inp,
                                           generator_numpy=rng_numpy)
        predicted_output = self.RNN.get_output()
        colors = ["r", "b", "g", "c", "m", "y", 'k']
        n_outputs = self.RNN.W_out.shape[0]
        for k in range(batch_size):
            for i in range(n_outputs):
                tag = labels[i] if not (labels is None) else ''
                axes[k].plot(predicted_output[i, :, k], color=colors[i], label=f'predicted {tag}')
                axes[k].plot(mask, target_batch[i, mask, k], color=colors[i], linestyle='--', label=f'target {tag}')
            axes[k].spines.right.set_visible(False)
            axes[k].spines.top.set_visible(False)

        axes[0].legend(fontsize=12, frameon=False, bbox_to_anchor=(1.0, 1.0))
        axes[batch_size // 2].set_ylabel("Output", fontsize=12)
        axes[-1].set_xlabel("time step, ms", fontsize=12)
        fig_output.tight_layout()
        plt.subplots_adjust(hspace=0.15, wspace=0.15)
        return fig_output


class PerformanceAnalyzerCDDM(PerformanceAnalyzer):
    def __init__(self, rnn_numpy):
        PerformanceAnalyzer.__init__(self, rnn_numpy)

    def calc_psychometric_data(self, task, mask, num_levels=7, num_repeats=7, sigma_rec=0.03, sigma_inp=0.03):
        coherence_lvls = np.linspace(-1, 1, num_levels)
        psychometric_data = {}
        psychometric_data["coherence_lvls"] = coherence_lvls
        psychometric_data["motion"] = {}
        psychometric_data["color"] = {}
        psychometric_data["motion"]["right_choice_percentage"] = np.empty((num_levels, num_levels))
        psychometric_data["color"]["right_choice_percentage"] = np.empty((num_levels, num_levels))
        psychometric_data["motion"]["MSE"] = np.empty((num_levels, num_levels))
        psychometric_data["color"]["MSE"] = np.empty((num_levels, num_levels))

        input_batch, target_batch, conditions = task.get_batch()
        batch_size = input_batch.shape[-1]
        input_batch = np.repeat(input_batch, axis=-1, repeats=num_repeats)
        target_batch = np.repeat(target_batch, axis=-1, repeats=num_repeats)
        self.RNN.clear_history()
        self.RNN.y = deepcopy(self.RNN.y_init)
        self.RNN.run(input_timeseries=input_batch,
                     sigma_rec=sigma_rec,
                     sigma_inp=sigma_inp,
                     save_history=True)
        output = self.RNN.get_output()
        if output.shape[0] == 2:
            choices = np.sign(output[0, -1, :] - output[1, -1, :])
        elif output.shape[0] == 1:
            choices = np.sign(output[0, -1, :])
        errors = np.sum(np.sum((target_batch[:, mask, :] - output[:, mask, :]) ** 2, axis=0), axis=0) / mask.shape[0]

        choices_to_right = (choices + 1) / 2
        # This reshaping pattern relies on the batch-structure from the CDDM task.
        # If you mess up with a batch generation function it may affect the psychometric function
        mean_choices_to_right = np.mean(choices_to_right.reshape(2, num_levels, num_levels, num_repeats), axis=-1)
        mean_error = np.mean(errors.reshape(2, num_levels, num_levels, num_repeats), axis=-1)
        psychometric_data["motion"]["right_choice_percentage"] = mean_choices_to_right[0, ...].T
        psychometric_data["color"]["right_choice_percentage"] = mean_choices_to_right[1, ...]
        psychometric_data["motion"]["MSE"] = mean_error[0, ...].T
        psychometric_data["color"]["MSE"] = mean_error[1, ...]
        self.psychometric_data = deepcopy(psychometric_data)
        return psychometric_data

    def plot_psychometric_data(self):
        coherence_lvls = self.psychometric_data["coherence_lvls"]
        Motion_rght_prcntg = self.psychometric_data["motion"]["right_choice_percentage"]
        Color_rght_prcntg = self.psychometric_data["color"]["right_choice_percentage"][::-1, :]
        Motion_MSE = self.psychometric_data["motion"]["MSE"]
        Color_MSE = self.psychometric_data["color"]["MSE"][::-1, :]
        num_lvls = Color_rght_prcntg.shape[0]

        fig, axes = plt.subplots(2, 2, figsize=(10, 8))
        fig.suptitle("Psychometric data", fontsize=14)
        tag = ["Motion", "Color"]
        for i in range(2):
            axes[0, i].title.set_text(f"{tag[i]}, % right")
            im1 = axes[0, i].imshow(eval(f"{tag[i]}_rght_prcntg"), cmap="coolwarm", interpolation="bicubic")
            plt.colorbar(im1, ax=axes[0, i], orientation='vertical')

            axes[1, i].title.set_text(f"{tag[i]}, MSE surface")
            im2 = axes[1, i].imshow(eval(f"{tag[i]}_MSE"), cmap="coolwarm", interpolation="bicubic")
            plt.colorbar(im2, ax=axes[1, i], orientation='vertical')

            for j in range(2):
                axes[j, i].set_xticks(np.arange(num_lvls), labels=np.round(coherence_lvls, 2), rotation=50)
                axes[j, i].set_yticks(np.arange(num_lvls), labels=np.round(coherence_lvls, 2)[::-1])
        axes[0, 0].set_xticklabels([])
        axes[0, 1].set_xticklabels([])
        axes[0, 0].set_ylabel("Coherence of color", fontsize=16)
        axes[1, 0].set_ylabel("Coherence of color", fontsize=16)
        axes[1, 0].set_xlabel("Coherence of motion", fontsize=16)
        axes[1, 1].set_xlabel("Coherence of motion", fontsize=16)
        fig.tight_layout()
        plt.subplots_adjust(wspace=0.125, hspace=0.15)
        return fig



import datajoint as dj

dj.config['database.host'] = 'localhost'
dj.config['database.user'] = 'ptolmachev'
dj.config['database.password'] = 'pawa'
# dj.config['database.host'] = 'datajoint-tengel.pni.princeton.edu'
# dj.config['database.user'] = 'pt1290'
# dj.config['database.password'] = 'a9Ab?spCKz$Zh@24h'
dj.config['display.limit'] = 500
dj.config["enable_python_native_blobs"] = True
dj.conn(reset=True)
schema = dj.schema('pt1290_RNNs')


@schema
class TaskDJ(dj.Manual):
    definition = """
    task_hash: char(16)
    ---
    task_name: char(16)  
    n_steps: smallint unsigned
    n_inputs: tinyint unsigned
    n_outputs: tinyint unsigned
    task_params : blob
    mask : blob
    """


@schema
class TrainerDJ(dj.Manual):
    definition = """
       trainer_hash: char(16)
       ---
       max_iter: int        
       tol: float
       lr: float
       lambda_orth: float
       lambda_r: float
       same_batch : tinyint
       shuffle : tinyint
    """


@schema
class RNNDJ(dj.Manual):
    definition = """
       -> TaskDJ
       -> TrainerDJ
       rnn_hash: char(16)                                                 # unique model id
       ---
       mse_score: float
       task_name: char(16) 
       timestamp: char(18)
       n: int                                                       # number of nodes in the RNN
       activation_name: enum('relu', 'tanh', 'sigmoid', 'softplus') # name of the activation function used in the dynamics
       constrained: tinyint                                         # boolean variable, either True or False, using biologically plausible connectivity or not
       dt: float                                                    # Euler integration timestep
       tau: float                                                   # Dynamical time-scale
       sr: float                                                    # spectral radius of the recurrent conenctivity
       connectivity_density_rec: Decimal(1,0)                       # opposite of sparsity of the connectivity. 1 - fully connected network
       sigma_rec: float
       sigma_inp: float
       w_inp : blob
       w_rec : blob
       w_out: blob
       b_rec: blob
       fp_data = NULL: mediumblob  
       psycho_data = NULL: mediumblob
       la_data = NULL: mediumblob  
    """



import os
import sys
sys.path.insert(0, os.getcwd())
sys.path.insert(0, '../')
import numpy as np
from scipy.sparse import random
from scipy.stats import uniform
from numpy.linalg import eig
from pathlib import Path
from matplotlib.colors import hsv_to_rgb
from matplotlib.colors import ListedColormap
import torch


def get_project_root():
    return Path(__file__).parent.parent


def numpify(function_torch):
    return lambda x: function_torch(torch.Tensor(x)).detach().numpy()


def in_the_list(x, x_list, diff_cutoff=1e-6):
    for i in range(len(x_list)):
        diff = np.linalg.norm(x - x_list[i], 2)
        if diff < diff_cutoff:
            return True
    return False


def orthonormalize(W):
    for i in range(W.shape[-1]):
        for j in range(i):
            W[:, i] = W[:, i] - W[:, j] * np.dot(W[:, i], W[:, j])
        W[:, i] = W[:, i] / np.linalg.norm(W[:, i])
    return W


def ReLU(x):
    return np.maximum(x, 0)


def generate_recurrent_weights(N, density, sr):
    A = (1.0 / (density * np.sqrt(N))) * np.array(random(N, N, density, data_rvs=uniform(-1, 2).rvs).todense())
    # get eigenvalues
    w, v = eig(A)
    A = A * (sr / np.max(np.abs(w)))
    return A


def sort_eigs(E, R):
    # sort eigenvectors
    data = np.hstack([E.reshape(-1, 1), R.T])
    data = np.array(sorted(data, key=lambda l: np.real(l[0])))[::-1, :]
    E = data[:, 0]
    R = data[:, 1:].T
    return E, R


def make_orientation_consistent(vectors, num_iter=10):
    vectors = np.stack(np.stack(vectors))
    for i in range(num_iter):  # np.min(dot_prod) < 0:
        average_vect = np.mean(vectors, axis=0)
        average_vect /= np.linalg.norm(average_vect)
        dot_prod = vectors @ average_vect
        vectors[np.where(dot_prod < 0)[0], :] *= -1
    return vectors


def cosine_sim(A, B):
    v1 = A.flatten() / np.linalg.norm(A.flatten())
    v2 = B.flatten() / np.linalg.norm(B.flatten())
    return np.round(np.dot(v1, v2), 3)


def get_colormaps():
    # define colors
    color1 = hsv_to_rgb([357.93 / 360, 84.06 / 100, 81.18 / 100])  # red
    color2 = hsv_to_rgb([226.73 / 360, 65 / 100, 64 / 100])  # blue
    color3 = hsv_to_rgb([246.16 / 360, 33 / 100, 74 / 100])  # bluish
    color4 = hsv_to_rgb([145.14 / 360, 93 / 100, 63 / 100])  # green
    color5 = hsv_to_rgb([28.32 / 360, 87.35 / 100, 96.08 / 100])  # orange
    color6 = hsv_to_rgb([196.11 / 360, 78 / 100, 93 / 100])  # light blue
    color7 = hsv_to_rgb([305 / 360, 53.66 / 100, 64.31 / 100])  # magenta

    colors = [color1, color2, color3, color4, color5, color6, color7]
    white = (1, 1, 1, 1)

    # for i in range(7):
    #     plt.plot(np.arange(9)-1*i, color=eval(f"color{i+1}"), linewidth = 10)
    # plt.show()

    newcolors = np.zeros((256, 4))
    newcolors[:, -1] = 1
    newcolors[:128, 0] = np.linspace(color2[0], white[0], 128)
    newcolors[128:, 0] = np.linspace(white[0], color1[0], 128)
    newcolors[:128, 1] = np.linspace(color2[1], white[1], 128)
    newcolors[128:, 1] = np.linspace(white[1], color1[1], 128)
    newcolors[:128, 2] = np.linspace(color2[2], white[2], 128)
    newcolors[128:, 2] = np.linspace(white[2], color1[2], 128)
    cmp_light = ListedColormap(newcolors)
    return colors, cmp_light



import json
import os
import sys

sys.path.insert(0, '../')
sys.path.insert(0, '../../')
from src.DynamicSystemAnalyzer import DynamicSystemAnalyzer
from src.PerformanceAnalyzer import PerformanceAnalyzer
from src.RNN_numpy import RNN_numpy
from src.utils import get_project_root, numpify, orthonormalize
from src.Trainer import Trainer
from src.RNN_torch import RNN_torch
from src.Task import *
from matplotlib import pyplot as plt
import torch
import time
from sklearn.decomposition import PCA

disp = True
activation = "tanh"
taskname = "MemoryAntiNumber"
train_config_file = f"train_config_{taskname}_{activation}.json"
config_dict = json.load(open(os.path.join(get_project_root(), "data", "configs", train_config_file), mode="r"))

# defining RNN:
N = config_dict["N"]
activation_name = config_dict["activation"]
if activation_name == 'relu':
    activation = lambda x: torch.maximum(x, torch.tensor(0))
elif activation_name == 'tanh':
    activation = torch.tanh
elif activation_name == 'sigmoid':
    activation = lambda x: 1 / (1 + torch.exp(-x))
elif activation_name == 'softplus':
    activation = lambda x: torch.log(1 + torch.exp(5 * x))

dt = config_dict["dt"]
tau = config_dict["tau"]
constrained = config_dict["constrained"]
connectivity_density_rec = config_dict["connectivity_density_rec"]
spectral_rad = config_dict["sr"]
sigma_inp = config_dict["sigma_inp"]
sigma_rec = config_dict["sigma_rec"]
seed = config_dict["seed"]
rng = torch.Generator()
if not seed is None:
    rng.manual_seed(seed)
input_size = config_dict["num_inputs"]
output_size = config_dict["num_outputs"]

# Task:
n_steps = config_dict["n_steps"]
task_params = config_dict["task_params"]

# Trainer:
lambda_orth = config_dict["lambda_orth"]
lambda_r = config_dict["lambda_r"]
mask = np.array(config_dict["mask"])
max_iter = config_dict["max_iter"]
tol = config_dict["tol"]
lr = config_dict["lr"]
weight_decay = config_dict["weight_decay"]
same_batch = config_dict["same_batch"]

# General:
tag = config_dict["tag"]
timestr = time.strftime("%Y%m%d-%H%M%S")
data_folder = os.path.join(config_dict["data_folder"], timestr)

# # creating instances:
rnn_torch = RNN_torch(N=N, dt=dt, tau=tau, input_size=input_size, output_size=output_size,
                      activation=activation, constrained=constrained,
                      sigma_inp=sigma_inp, sigma_rec=sigma_rec,
                      connectivity_density_rec=connectivity_density_rec,
                      spectral_rad=spectral_rad,
                      random_generator=rng)
task = eval("Task" + taskname)(n_steps=n_steps, n_inputs=input_size, n_outputs=output_size, task_params=task_params)
criterion = torch.nn.MSELoss()
optimizer = torch.optim.Adam(rnn_torch.parameters(),
                             lr=lr,
                             weight_decay=weight_decay)
trainer = Trainer(RNN=rnn_torch, Task=task,
                  max_iter=max_iter, tol=tol,
                  optimizer=optimizer, criterion=criterion,
                  lambda_orth=lambda_orth, lambda_r=lambda_r)

# datasaver = DataSaver(data_folder)
datasaver = None
try:
    SLURM_JOB_ID = int(os.environ["SLURM_JOB_ID"])
    task_params["seed"] = SLURM_JOB_ID
    seed = SLURM_JOB_ID
except:
    SLURM_JOB_ID = None

rnn_trained, train_losses, val_losses, net_params = trainer.run_training(train_mask=mask, same_batch=same_batch)
fig_trainloss = plt.figure(figsize=(10, 3))
plt.plot(train_losses, color='r', label='train loss (log scale)')
plt.plot(val_losses, color='b', label='valid loss (log scale)')
plt.yscale("log")
plt.grid(True)
plt.legend(fontsize=16)
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_trainloss, "train&valid_loss")

# best_net_params = pickle.load(open(os.path.join(get_project_root(), "data", "trained_RNNs", "MemoryAntiNumber", "20230122-220547", "params_MemoryAntiNumber_0.02719.pkl"), "rb+"))

# validate
RNN_valid = RNN_numpy(N=net_params["N"],
                      dt=net_params["dt"],
                      tau=net_params["tau"],
                      activation=numpify(activation),
                      W_inp=net_params["W_inp"],
                      W_rec=net_params["W_rec"],
                      W_out=net_params["W_out"],
                      bias_rec=net_params["bias_rec"],
                      y_init=net_params["y_init"])

analyzer = PerformanceAnalyzer(RNN_valid)
score_function = lambda x, y: np.mean((x - y) ** 2)
input_batch_valid, target_batch_valid, conditions_valid = task.get_batch()
score = analyzer.get_validation_score(score_function, input_batch_valid, target_batch_valid,
                                      mask, sigma_rec=sigma_rec, sigma_inp=sigma_inp)
print(f"MSE validation: {np.round(score, 5)}")
if not (datasaver is None): datasaver.save_data(config_dict, "config.json")
if not (datasaver is None): datasaver.save_data(net_params, f"params_{taskname}_{np.round(score, 5)}.pkl")

print(f"Plotting random trials")
inds = np.random.choice(np.arange(input_batch_valid.shape[-1]), 12)
inputs = input_batch_valid[..., inds]
targets = target_batch_valid[..., inds]

fig_trials = analyzer.plot_trials(inputs, targets, mask, sigma_rec=sigma_rec, sigma_inp=sigma_inp)
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_trials, "random_trials")

dsa = DynamicSystemAnalyzer(RNN_valid)
params = {"fun_tol": 0.05,
          "diff_cutoff": 1e-4,
          "sigma_init_guess": 15,
          "patience": 100,
          "stop_length": 100,
          "mode": "approx"}
dsa.get_fixed_points(Input=np.array([0, 0]), **params)
dsa.get_fixed_points(Input=np.array([0, 1]), **params)

P = np.zeros((RNN_valid.N, 2))
P[:, 0] = RNN_valid.W_out
input_as_key = str([0, 0])
points = dsa.fp_data[input_as_key]
all_points = np.vstack([points[type] for type in list(points.keys())])
pca = PCA(n_components=1)
pca.fit(all_points)
P[:, 1] = pca.components_.flatten()
P = orthonormalize(P)

fig_fp = dsa.plot_fixed_points(projection='2D', P=P)
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_fp, "slow_points_projection_output")

fig_fp = dsa.plot_fixed_points(projection='3D')
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_fp, "slow_points_projection_3D_PCA")

# fig_fp = dsa.plot_fixed_points(projection='3D')
# if disp:
#     plt.show()
# if not (datasaver is None): datasaver.save_figure(fig_fp, "slow_points_projection_3D_PCA")



import json
import os
import pickle
from datetime import date

class DataSaver():
    '''
    Class which encapsulates creating data folders and saving information there afterwards
    '''

    def __init__(self, data_folder, dj_integration=False):
        # create data folder if doesn't exist
        os.umask(0)
        os.makedirs(data_folder, exist_ok=True, mode=0o777)
        self.data_folder = data_folder
        self.date_tag = ''.join((list(str(date.today()).split("-"))[::-1]))
        self.dj_integration = dj_integration

    def save_data(self, data, file_name):
        '''save data as a pickle or json file, depending on the name'''
        if 'pkl' in file_name:
            pickle.dump(data, open(os.path.join(self.data_folder, file_name), "wb+"))
        elif 'json' in file_name:
            json_obj = json.dumps(data, indent=4)
            outfile = open(os.path.join(self.data_folder, file_name), mode="w")
            outfile.write(json_obj)
        return None

    def save_figure(self, figure, file_name):
        '''saving an image as a png'''
        figure.savefig(os.path.join(self.data_folder, file_name), dpi=300, format='png')
        return None

    # def save_to_dj(self, data_dict):



from copy import deepcopy

import numpy as np

'''
Class which generates the input time-series and the correct output for the CDDM task for multiple coherences
'''


class Task():
    def __init__(self, n_steps, n_inputs, n_outputs, task_params):
        '''
        :param n_inputs: number of input channels
        :param num_outputs: number of target output-time series.
        '''
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.task_params = task_params
        self.seed = task_params["seed"]
        if not (self.seed is None):
            self.rng = np.random.default_rng(seed=self.seed)
        else:
            self.rng = np.random.default_rng()

    def generate_input_target_stream(self, **kwargs):
        '''
        input_stream should have a dimensionality n_inputs x n_steps
        target_stream should have a dimensionality n_outputs x n_steps
        :param kwargs:
        :return:
        '''
        raise NotImplementedError("This is a generic Task class!")

    def get_batch(self, **kwargs):
        raise NotImplementedError("This is a generic Task class!")


class TaskCDDM(Task):

    def __init__(self, n_steps, n_inputs, n_outputs, task_params):
        '''
        :param n_steps: number of steps in the trial, default is 750
        '''
        Task.__init__(self, n_steps, n_inputs, n_outputs, task_params)
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.cue_on = self.task_params["cue_on"]
        self.cue_off = self.task_params["cue_off"]
        self.stim_on = self.task_params["stim_on"]
        self.stim_off = self.task_params["stim_off"]
        self.dec_on = self.task_params["dec_on"]
        self.dec_off = self.task_params["dec_off"]
        self.coherences = self.task_params["coherences"]

    def generate_input_target_stream(self, context, motion_coh, color_coh):
        '''
        generate an input and target for a single trial with the supplied coherences
        :param context: could be either 'motion' or 'color' (see Mante et. all 2013 paper)
        :param motion_coh: coherence of information in motion channel, range: (0, 1)
        :param color_coh: coherence of information in color channel, range: (0, 1)
        :return: input_stream, target_stream
        input_stream - input time series (both context and sensory): n_inputs x num_steps
        target_stream - time sereis reflecting the correct decision: num_outputs x num_steps

        :param protocol_dict: a dictionary which provides the trial structure:
        cue_on, cue_off - defines the timespan when the contextual information is supplied
        stim_on, stim_off - defines the timespan when the sensory information is supplied
        dec_on, dec_off - defines the timespan when the decision has to be present in the target stream
        all the values should be less than n_steps
        '''

        # given the context and coherences of signals
        # generate input array (n_inputs, n_steps)
        # and target array (ideal output of the Decision-making system)

        # Transform coherence to signal
        motion_r = (1 + motion_coh) / 2
        motion_l = 1 - motion_r
        color_r = (1 + color_coh) / 2
        color_l = 1 - color_r

        # Cue input stream
        cue_input = np.zeros((self.n_inputs, self.n_steps))
        ind_ctxt = 0 if context == "motion" else 1
        cue_input[ind_ctxt, self.cue_on:self.cue_off] = np.ones(self.cue_off - self.cue_on)

        sensory_input = np.zeros((self.n_inputs, self.n_steps))
        # Motion input stream
        sensory_input[2, self.stim_on - 1:self.stim_off] = motion_r * np.ones([self.stim_off - self.stim_on + 1])
        sensory_input[3, self.stim_on - 1:self.stim_off] = motion_l * np.ones([self.stim_off - self.stim_on + 1])
        # Color input stream
        sensory_input[4, self.stim_on - 1:self.stim_off] = color_r * np.ones([self.stim_off - self.stim_on + 1])
        sensory_input[5, self.stim_on - 1:self.stim_off] = color_l * np.ones([self.stim_off - self.stim_on + 1])
        input_stream = cue_input + sensory_input

        # Target stream
        if self.n_outputs == 1:
            target_stream = np.zeros((1, self.n_steps))
            target_stream[0, self.dec_on - 1:self.dec_off] = np.sign(motion_coh) if (context == 'motion') else np.sign(
                color_coh)
        elif self.n_outputs == 2:
            target_stream = np.zeros((2, self.n_steps))
            relevant_coh = motion_coh if (context == 'motion') else color_coh
            if relevant_coh == 0.0:
                pass
            else:
                decision = np.sign(relevant_coh)
                ind = 0 if (decision == 1.0) else 1
                target_stream[ind, self.dec_on - 1:self.dec_off] = 1
        return input_stream, target_stream

    def get_batch(self, shuffle=False):
        '''
        coherences: list containing range of coherences for each channel (e.g. [-1, -0.5, -0.25,  0, 0.25, 0.5, 1]
        :param shuffle: shuffle the final array
        :param generator_numpy: the random generator (for reproducibility, if using shuffle=True)
        :return: array of inputs, array of targets, and the conditions (context, coherences and the correct choice)
        '''
        coherences = self.task_params["coherences"]
        inputs = []
        targets = []
        conditions = []
        if self.rng is None:
            generator_numpy = np.random.default_rng()
        for context in ["motion", "color"]:
            for c1 in coherences:
                for c2 in coherences:
                    relevant_coh = c1 if context == 'motion' else c2
                    irrelevant_coh = c2 if context == 'motion' else c1
                    motion_coh = c1 if context == 'motion' else c2
                    color_coh = c1 if context == 'color' else c2
                    coh_pair = (relevant_coh, irrelevant_coh)

                    correct_choice = 1 if ((context == "motion" and motion_coh > 0) or (
                            context == "color" and color_coh > 0)) else -1
                    conditions.append({'context': context,
                                       'motion_coh': motion_coh,
                                       'color_coh': color_coh,
                                       'correct_choice': correct_choice})
                    input_stream, target_stream = self.generate_input_target_stream(context, coh_pair[0], coh_pair[1])
                    inputs.append(deepcopy(input_stream))
                    targets.append(deepcopy(target_stream))

        # batch_size should be a last dimension
        inputs = np.stack(inputs, axis=2)
        targets = np.stack(targets, axis=2)
        if shuffle:
            perm = self.rng.permutation(len(inputs))
            inputs = inputs[..., perm]
            targets = targets[..., perm]
            conditions = [conditions[index] for index in perm]
        return inputs, targets, conditions


class TaskDMTS(Task):

    def __init__(self, n_steps, n_inputs, task_params, n_outputs=2):
        '''
        :param n_steps: number of steps in the trial, default is 750
        '''
        Task.__init__(self, n_steps, n_inputs, n_outputs, task_params)
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.stim_on_sample = self.task_params["stim_on_sample"]
        self.stim_off_sample = self.task_params["stim_off_sample"]
        self.stim_on_match = self.task_params["stim_on_match"]
        self.stim_off_match = self.task_params["stim_off_match"]
        self.dec_on = self.task_params["dec_on"]
        self.dec_off = self.task_params["dec_off"]

    def generate_input_target_stream(self, num_sample_channel, num_match_channel):
        input_stream = np.zeros([self.n_inputs, self.n_steps])
        input_stream[num_sample_channel, self.stim_on_sample:self.stim_off_sample] = 1.0
        input_stream[num_match_channel, self.stim_on_match:self.stim_off_match] = 1.0

        # Target stream
        target_stream = np.zeros((2, self.n_steps))
        if (num_sample_channel == num_match_channel):
            target_stream[0, self.dec_on: self.dec_off] = 1
        elif (num_sample_channel != num_match_channel):
            target_stream[1, self.dec_on: self.dec_off] = 1
        return input_stream, target_stream

    def get_batch(self, shuffle=False):
        # batch size = 256 for two inputs
        inputs = []
        targets = []
        conditions = []
        for num_sample_channel in range(self.n_inputs):
            for num_match_channel in range(self.n_inputs):
                correct_choice = 1 if (num_sample_channel == num_match_channel) else -1
                conditions.append({'num_sample_channel': num_sample_channel,
                                   'num_match_channel': num_match_channel,
                                   'correct_choice': correct_choice})
                input_stream, target_stream = self.generate_input_target_stream(num_sample_channel, num_match_channel)
                inputs.append(deepcopy(input_stream))
                targets.append(deepcopy(target_stream))

        inputs = 64 * inputs
        targets = 64 * targets
        conditions = 64 * conditions

        inputs = np.stack(inputs, axis=2)
        targets = np.stack(targets, axis=2)
        if shuffle:
            perm = self.rng.permutation(len(inputs))
            inputs = inputs[..., perm]
            targets = targets[..., perm]
            conditions = [conditions[index] for index in perm]
        return inputs, targets, conditions


class TaskNBitFlipFlop(Task):
    def __init__(self, n_steps, n_inputs, n_outputs, task_params):
        '''
        for tanh neurons only
        '''
        Task.__init__(self, n_steps, n_inputs, n_outputs, task_params)
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.mu = self.task_params["mu"]
        self.n_refractory = self.n_flip = self.task_params["n_flip_steps"]
        self.lmbd = self.mu / self.n_steps

    def generate_flipflop_times(self):
        inds = []
        last_ind = 0
        while last_ind < self.n_steps:
            r = np.random.rand()
            ind = last_ind + self.n_refractory + int(-(1 / self.lmbd) * np.log(r))
            if (ind < self.n_steps): inds.append(ind)
            last_ind = ind
        return inds

    def generate_input_target_stream(self):
        input_stream = np.zeros((self.n_inputs, self.n_steps))
        target_stream = np.zeros((self.n_outputs, self.n_steps))
        condition = {}
        for n in range(self.n_inputs):
            inds_flips_and_flops = self.generate_flipflop_times()
            mask = [0 if np.random.rand() < 0.5 else 1 for i in range(len(inds_flips_and_flops))]
            inds_flips = []
            inds_flops = []
            for i in range(len(inds_flips_and_flops)):
                if mask[i] == 0:
                    inds_flops.append(inds_flips_and_flops[i])
                elif mask[i] == 1.0:
                    inds_flips.append(inds_flips_and_flops[i])
            for ind in inds_flips:
                input_stream[n, ind: ind + self.n_refractory] = 1.0
            for ind in inds_flops:
                input_stream[n, ind: ind + self.n_refractory] = -1.0

            last_flip_ind = 0
            last_flop_ind = 0
            for i in range(self.n_steps):
                if i in inds_flips:
                    last_flip_ind = i
                elif i in inds_flops:
                    last_flop_ind = i
                if last_flop_ind < last_flip_ind:
                    target_stream[n, i] = 1.0
                elif last_flop_ind > last_flip_ind:
                    target_stream[n, i] = -1.0
            condition[n] = {"inds_flips": inds_flips, "inds_flops": inds_flops}
        return input_stream, target_stream, condition

    def get_batch(self, batch_size=256, shuffle=False, generator_numpy=None):
        inputs = []
        targets = []
        conditions = []
        for i in range(batch_size):
            input_stream, target_stream, condition = self.generate_input_target_stream()
            inputs.append(deepcopy(input_stream))
            targets.append(deepcopy(target_stream))
            conditions.append(deepcopy(condition))
        inputs = np.stack(inputs, axis=2)
        targets = np.stack(targets, axis=2)
        if shuffle:
            perm = self.rng.permutation(len(inputs))
            inputs = inputs[..., perm]
            targets = targets[..., perm]
            conditions = [conditions[index] for index in perm]
        return inputs, targets, conditions


class TaskMemoryAntiNumber(Task):
    def __init__(self, n_steps, n_inputs, n_outputs, task_params):
        '''
        Given an one-channel input x in range (-2, 2) for a short period of time
        Output -x after in the `recall' period signified by an additional input +1 in the second input channel.

        '''
        Task.__init__(self, n_steps, n_inputs, n_outputs, task_params)
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.stim_on_range = task_params["stim_on_range"]
        self.stim_duration = task_params["stim_duration"]
        self.recall_on = task_params["recall_on"]
        self.recall_off = task_params["recall_off"]

    def generate_input_target_stream(self, number):
        stim_on = int(self.rng.uniform(*self.stim_on_range))
        duration = self.stim_duration
        input_stream = np.zeros((self.n_inputs, self.n_steps))
        target_stream = np.zeros((self.n_outputs, self.n_steps))
        input_stream[0, stim_on: stim_on + duration] = number
        input_stream[1, self.recall_on: self.recall_off] = 1

        target_stream[0, self.recall_on: self.recall_off] = -number
        condition = {"number": number}
        return input_stream, target_stream, condition

    def get_batch(self, shuffle=False):
        inputs = []
        targets = []
        conditions = []
        numbers = np.linspace(-2, 2, 32)
        for number in numbers:
            input_stream, target_stream, condition = self.generate_input_target_stream(number)
            inputs.append(deepcopy(input_stream))
            targets.append(deepcopy(target_stream))
            conditions.append(deepcopy(condition))
        inputs = np.stack(inputs, axis=2)
        targets = np.stack(targets, axis=2)
        inputs = np.repeat(inputs, axis=2, repeats=11)
        targets = np.repeat(targets, axis=2, repeats=11)
        if shuffle:
            perm = self.rng.permutation(len(inputs))
            inputs = inputs[..., perm]
            targets = targets[..., perm]
            conditions = [conditions[index] for index in perm]
        return inputs, targets, conditions


class TaskMemoryAntiAngle(Task):
    def __init__(self, n_steps, n_inputs, n_outputs, task_params):
        '''
        Given a two-channel input 2 cos(theta) and 2 sin(theta) specifying an angle theta (present only for a short period of time),
        Output Acos(theta+pi), Asin(theta+pi) in the recall period (signified by +1 provided in the third input-channel)
        This task is similar (but not exactly the same) to the task described in
        "Flexible multitask computation in recurrent networks utilizes shared dynamical motifs"
        Laura Driscoll1, Krishna Shenoy, David Sussillo
        '''
        Task.__init__(self, n_steps, n_inputs, n_outputs, task_params)
        self.n_steps = n_steps
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.stim_on_range = task_params["stim_on_range"]
        self.stim_duration = task_params["stim_duration"]
        self.recall_on = task_params["recall_on"]
        self.recall_off = task_params["recall_off"]

    def generate_input_target_stream(self, theta):
        input_stream = np.zeros((self.n_inputs, self.n_steps))
        target_stream = np.zeros((self.n_outputs, self.n_steps))
        stim_on = int(self.rng.uniform(*self.stim_on_range))
        duration = self.stim_duration

        input_stream[0, stim_on: stim_on + duration] = 2 * np.cos(theta)
        input_stream[1, stim_on: stim_on + duration] = 2 * np.sin(theta)
        input_stream[2, self.recall_on: self.recall_off] = 1

        # Supplying it with an explicit instruction to recall the theta + 180
        target_stream[0, self.recall_on: self.recall_off] = 2 * np.cos(theta + np.pi)
        target_stream[1, self.recall_on: self.recall_off] = 2 * np.sin(theta + np.pi)
        condition = {"theta": theta}
        return input_stream, target_stream, condition

    def get_batch(self, shuffle=False):
        inputs = []
        targets = []
        conditions = []
        thetas = 2 * np.pi * np.linspace(0, 1, 37)[:-1]
        for theta in thetas:
            input_stream, target_stream, condition = self.generate_input_target_stream(theta)
            inputs.append(deepcopy(input_stream))
            targets.append(deepcopy(target_stream))
            conditions.append(deepcopy(condition))
        inputs = np.stack(inputs, axis=2)
        targets = np.stack(targets, axis=2)
        inputs = np.repeat(inputs, axis=2, repeats=11)
        targets = np.repeat(targets, axis=2, repeats=11)
        if shuffle:
            perm = self.rng.permutation(len(inputs))
            inputs = inputs[..., perm]
            targets = targets[..., perm]
            conditions = [conditions[index] for index in perm]
        return inputs, targets, conditions


if __name__ == '__main__':
    # n_steps = 750
    # n_inputs = 6
    # n_outputs = 2
    # task_params = dict()
    # task_params["coherences"] = [-1, -0.5, -0.25, 0, 0.25, 0.5, 1]
    # task_params["cue_on"] = 0
    # task_params["cue_off"] = 750
    # task_params["stim_on"] = 250
    # task_params["stim_off"] = 750
    # task_params["dec_on"] = 500
    # task_params["dec_off"] = 750
    # task = TaskCDDM(n_steps, n_inputs, n_outputs, task_params)
    # inputs, targets, conditions = task.get_batch()
    # print(inputs.shape, targets.shape)

    # n_steps = 750
    # n_inputs = 2
    # n_outputs = 2
    # task_params = dict()
    # task_params["n_steps"] = n_steps
    # task_params["n_inputs"] = n_inputs
    # task_params["n_outputs"] = n_outputs
    # task_params["stim_on_sample"] = 100
    # task_params["stim_off_sample"] = 200
    # task_params["stim_on_match"] = 300
    # task_params["stim_off_match"] = 400
    # task_params["dec_on"] = 500
    # task_params["dec_off"] = 750
    # task = TaskDMTS(n_steps, n_inputs, task_params)
    # inputs, targets, conditions = task.get_batch()
    # print(inputs.shape, targets.shape)
    #
    #
    # n_steps = 750
    # n_inputs = 2
    # n_outputs = 2
    # task_params = dict()
    # task_params["n_steps"] = n_steps
    # task_params["n_inputs"] = n_inputs
    # task_params["n_outputs"] = n_outputs
    # task_params["n_flip_steps"] = 20
    # task_params["mu"] = 7
    # task = TaskNBitFlipFlop(n_steps, n_inputs, n_outputs, task_params)
    # inputs, targets, conditions = task.get_batch()
    # print(inputs.shape, targets.shape)

    # n_steps = 320
    # n_inputs = 3
    # n_outputs = 2
    # task_params = dict()
    # task_params["stim_on"] = n_steps // 8
    # task_params["stim_off"] = 3 * n_steps//16
    # task_params["recall_on"] = 5 * n_steps//8
    # task_params["recall_off"] = n_steps
    # task = TaskMemoryAntiAngle(n_steps, n_inputs, n_outputs, task_params)
    # inputs, targets, conditions = task.get_batch()
    # print(inputs.shape, targets.shape)

    n_steps = 320
    n_inputs = 2
    n_outputs = 1
    task_params = dict()
    task_params["stim_on"] = n_steps // 8
    task_params["stim_off"] = 3 * n_steps // 16
    task_params["recall_on"] = 5 * n_steps // 8
    task_params["recall_off"] = n_steps
    task = TaskMemoryAntiNumber(n_steps, n_inputs, n_outputs, task_params)
    inputs, targets, conditions = task.get_batch()
    print(inputs.shape, targets.shape)



from setuptools import setup, find_packages

setup(name='rnn_coach',
      version='0.1',
      description='training RNNs with various activation functions on user-defined tasks. Anaylzing them with fixed and slow points analysis.',
      url='https://github.com/engellab/rnn-coach',
      author='Pavel Tolmachev',
      author_email='pt1290@princeton.edu',
      license='MIT',
      packages=find_packages(),
      zip_safe=False)


import json
import os
import sys

sys.path.insert(0, '../')
sys.path.insert(0, '../../')
from src.DataSaver import DataSaver
from src.DynamicSystemAnalyzer import DynamicSystemAnalyzerCDDM
from src.PerformanceAnalyzer import PerformanceAnalyzerCDDM
from src.RNN_numpy import RNN_numpy
from src.utils import get_project_root, numpify
from src.Trainer import Trainer
from src.RNN_torch import RNN_torch
from src.Task import *
from matplotlib import pyplot as plt
import torch
import time

# from src.datajoint_config import *

disp = True
activation = "relu"
taskname = "CDDM"
train_config_file = f"train_config_{taskname}_{activation}.json"
config_dict = json.load(
    open(os.path.join(get_project_root(), "data", "configs", train_config_file), mode="r", encoding='utf-8'))

# defining RNN:
N = config_dict["N"]
activation_name = config_dict["activation"]
if activation_name == 'relu':
    activation = lambda x: torch.maximum(x, torch.tensor(0))
elif activation_name == 'tanh':
    activation = torch.tanh
elif activation_name == 'sigmoid':
    activation = lambda x: 1 / (1 + torch.exp(-x))
elif activation_name == 'softplus':
    activation = lambda x: torch.log(1 + torch.exp(5 * x))
dt = config_dict["dt"]
tau = config_dict["tau"]
constrained = config_dict["constrained"]
connectivity_density_rec = config_dict["connectivity_density_rec"]
spectral_rad = config_dict["sr"]
sigma_inp = config_dict["sigma_inp"]
sigma_rec = config_dict["sigma_rec"]
seed = config_dict["seed"]

rng = torch.Generator()
if not seed is None:
    rng.manual_seed(seed)
input_size = config_dict["num_inputs"]
output_size = config_dict["num_outputs"]

# Task:
n_steps = config_dict["n_steps"]
task_params = config_dict["task_params"]

# Trainer:
lambda_orth = config_dict["lambda_orth"]
lambda_r = config_dict["lambda_r"]
mask = np.array(config_dict["mask"])
max_iter = config_dict["max_iter"]
tol = config_dict["tol"]
lr = config_dict["lr"]
weight_decay = config_dict["weight_decay"]
same_batch = config_dict["same_batch"]

# General:
tag = config_dict["tag"]
timestr = time.strftime("%Y%m%d-%H%M%S")
# if run on the cluster
try:
    SLURM_JOB_ID = int(os.environ["SLURM_ARRAY_TASK_ID"])
    task_params["seed"] = np.random.randint(1000000)
    seed = task_params["seed"]
except:
    SLURM_JOB_ID = None

data_folder = os.path.join(config_dict["data_folder"], timestr + ('' if (SLURM_JOB_ID is None) else str(SLURM_JOB_ID)))

# # creating instances:
rnn_torch = RNN_torch(N=N, dt=dt, tau=tau, input_size=input_size, output_size=output_size,
                      activation=activation, constrained=constrained,
                      sigma_inp=sigma_inp, sigma_rec=sigma_rec,
                      connectivity_density_rec=connectivity_density_rec,
                      spectral_rad=spectral_rad,
                      random_generator=rng)
task = TaskCDDM(n_steps=n_steps, n_inputs=input_size, n_outputs=output_size, task_params=task_params)
criterion = torch.nn.MSELoss()
optimizer = torch.optim.Adam(rnn_torch.parameters(),
                             lr=lr,
                             weight_decay=weight_decay)
trainer = Trainer(RNN=rnn_torch, Task=task,
                  max_iter=max_iter, tol=tol,
                  optimizer=optimizer, criterion=criterion,
                  lambda_orth=lambda_orth, lambda_r=lambda_r)

tic = time.perf_counter()
rnn_trained, train_losses, val_losses, net_params = trainer.run_training(train_mask=mask, same_batch=same_batch)
toc = time.perf_counter()
print(f"Executed training in {toc - tic:0.4f} seconds")

# net_params = pickle.load(open(os.path.join(get_project_root(), "data", "trained_RNNs", "CDDM", "20230117-175732", "params_CDDM_0.03556.pkl"), "rb+"))
# validate
RNN_valid = RNN_numpy(N=net_params["N"],
                      dt=net_params["dt"],
                      tau=net_params["tau"],
                      activation=numpify(activation),
                      W_inp=net_params["W_inp"],
                      W_rec=net_params["W_rec"],
                      W_out=net_params["W_out"],
                      bias_rec=net_params["bias_rec"],
                      y_init=net_params["y_init"])

analyzer = PerformanceAnalyzerCDDM(RNN_valid)
score_function = lambda x, y: np.mean((x - y) ** 2)
input_batch_valid, target_batch_valid, conditions_valid = task.get_batch()
score = analyzer.get_validation_score(score_function, input_batch_valid, target_batch_valid,
                                      mask, sigma_rec=sigma_rec, sigma_inp=sigma_inp)
score = np.round(score, 7)
datasaver = DataSaver(data_folder)
# datasaver = None

print(f"MSE validation: {score}")
if not (datasaver is None): datasaver.save_data(config_dict, f"{score}_config.json")
if not (datasaver is None): datasaver.save_data(net_params, f"params_{taskname}_{score}.pkl")

fig_trainloss = plt.figure(figsize=(10, 3))
plt.plot(train_losses, color='r', label='train loss (log scale)')
plt.plot(val_losses, color='b', label='valid loss (log scale)')
plt.yscale("log")
plt.grid(True)
plt.legend(fontsize=16)
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_trainloss, f"{score}_train&valid_loss.png")

print(f"Plotting random trials")
inds = np.random.choice(np.arange(input_batch_valid.shape[-1]), 12)
inputs = input_batch_valid[..., inds]
targets = target_batch_valid[..., inds]

fig_trials = analyzer.plot_trials(inputs, targets, mask, sigma_rec=sigma_rec, sigma_inp=sigma_inp)
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_trials, f"{score}_random_trials.png")

print(f"Plotting psychometric data")
num_levels = len(config_dict["task_params"]["coherences"])
analyzer.calc_psychometric_data(task, mask, num_levels=num_levels, num_repeats=31, sigma_rec=0.03, sigma_inp=0.03)
fig_psycho = analyzer.plot_psychometric_data()
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_psycho, f"{score}_psychometric_data.png")
if not (datasaver is None): datasaver.save_data(analyzer.psychometric_data, f"{score}_psycho_data.pkl")

print(f"Analyzing fixed points")
dsa = DynamicSystemAnalyzerCDDM(RNN_valid)
params = {"fun_tol": 0.05,
          "diff_cutoff": 1e-4,
          "sigma_init_guess": 15,
          "patience": 100,
          "stop_length": 100,
          "mode": "approx"}
dsa.get_fixed_points(Input=np.array([1, 0, 0.5, 0.5, 0.5, 0.5]), **params)
dsa.get_fixed_points(Input=np.array([0, 1, 0.5, 0.5, 0.5, 0.5]), **params)
print(f"Calculating Line Attractor analytics")
dsa.calc_LineAttractor_analytics()

fig_LA3D = dsa.plot_LineAttractor_3D()
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_LA3D, f"{score}_LA_3D.png")
if not (datasaver is None): datasaver.save_data(dsa.fp_data, f"{score}_fp_data.pkl")
if not (datasaver is None): datasaver.save_data(dsa.LA_data, f"{score}_LA_data.pkl")

fig_RHS = dsa.plot_RHS_over_LA()
if disp:
    plt.show()
if not (datasaver is None): datasaver.save_figure(fig_RHS, f"{score}_LA_RHS.png")

# rnn_dj = RNNDJ()
# task_dj = TaskDJ()
# trainer_dj = TrainerDJ()
# cddm_analysis_dj = CDDMRNNAnalysisDJ()
#
# task_id = 0
# trainer_id = 0
# rnn_timestamp = time.strftime("%Y%m%d%H%M%S")
#
# task_dj_dict = {"task_name": taskname + "_" + str(task_id),
#                 "n_steps": config_dict["n_steps"],
#                 "n_inputs": config_dict["num_inputs"],
#                 "n_outputs": config_dict["num_outputs"],
#                 "task_params": config_dict["task_params"],
#                 "mask": mask}
# trainer_dj_dict = {"task_name": taskname + "_" + str(task_id),
#                    "trainer_id": trainer_id,
#                    "max_iter": config_dict["max_iter"],
#                    "tol": config_dict["tol"],
#                    "lr": config_dict["lr"],
#                    "lambda_orth": config_dict["lambda_orth"],
#                    "lambda_r": config_dict["lambda_r"],
#                    "same_batch" : config_dict["same_batch"],
#                    "shuffle" : False}
# rnn_dj_dict = {"task_name": taskname + "_" + str(task_id),
#                "rnn_timestamp" : rnn_timestamp,
#                "trainer_id" : trainer_id,
#                "n": config_dict["N"],
#                "activation_name": config_dict["activation"],
#                "constrained": config_dict["constrained"],
#                "dt": config_dict["dt"],
#                "tau": config_dict["tau"],
#                "sr": config_dict["sr"],
#                "connectivity_density_rec": config_dict["connectivity_density_rec"],
#                "sigma_rec" : config_dict["sigma_rec"],
#                "sigma_inp": config_dict["sigma_inp"],
#                "w_inp" : net_params["W_inp"],
#                "w_rec" : net_params["W_rec"],
#                "w_out" : net_params["W_out"],
#                "b_rec" : 0 if net_params["bias_rec"] is None else net_params["bias_rec"]}
# cddm_analysis_dj_dict = {"task_name": taskname + "_" + str(task_id),
#                          "rnn_timestamp" : rnn_timestamp,
#                          "trainer_id": trainer_id,
#                          "mse_score": score,
#                          "psycho_data": deepcopy(analyzer.psychometric_data),
#                          "fp_data": deepcopy(dsa.fp_data),
#                          "la_data": deepcopy(dsa.LA_data)}
#
# task_dj.insert1(task_dj_dict, skip_duplicates=True)
# rnn_dj.insert1(rnn_dj_dict, skip_duplicates=True)
# trainer_dj.insert1(trainer_dj_dict, skip_duplicates=True)
# trainer_dj.insert1(trainer_dj_dict, skip_duplicates=True)



import sys
sys.path.insert(0, "../")
from copy import deepcopy
import torch
import numpy as np
from numpy import linalg


# Connectivity defining methods

def sparse(tensor, sparsity, mean=0, std=1, generator=None):
    r"""Fills the 2D input `Tensor` as a sparse matrix, where the
    non-zero elements will be drawn from the normal distribution
    :math:`\mathcal{N}(0, 0.01)`, as described in `Deep learning via
    Hessian-free optimization` - Martens, J. (2010).

    Args:
        tensor: an n-dimensional `torch.Tensor`
        sparsity: The fraction of elements in each column to be set to zero
        std: the standard deviation of the normal distribution used to generate
            the non-zero values

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.sparse_(w, sparsity=0.1)
    """
    if tensor.ndimension() != 2:
        raise ValueError("Only tensors with 2 dimensions are supported")

    rows, cols = tensor.shape
    num_zeros = int(np.ceil(sparsity * rows))

    with torch.no_grad():
        tensor.normal_(mean, std, generator=generator)
        for col_idx in range(cols):
            row_indices = torch.randperm(rows, generator=generator)
            zero_indices = row_indices[:num_zeros]
            tensor[zero_indices, col_idx] = 0
    return tensor


def get_connectivity(device, N, num_inputs, num_outputs, radius=1.5, recurrent_density=1, input_density=1,
                     output_density=1, generator=None):
    '''
    generates W_inp, W_rec and W_out matrices of RNN, with specified parameters
    :param device: torch related: CPU or GPU
    :param N: number of neural nodes
    :param num_inputs: number of input channels, input dimension
    :param num_outputs: number of output channels, output dimension
    :param radius: spectral radius of the generated cnnectivity matrix: controls the maximal abs value of eigenvectors.
    the greater the parameter is the more sustained and chaotic activity the network exchibits, the lower - the quicker
    the network relaxes back to zero.
    :param recurrent_density: oppposite of sparcirty of the reccurrent matrix. 1.0 - fully connected recurrent matrix
    :param input_density: 1.0 - fully connected input matrix, 0 - maximally sparce matrix
    :param output_density: 1.0 - fully connected output matrix, 0 - maximally sparce matrix
    :param generator: torch random generator, for reproducibility
    :return:
    '''

    # Balancing parameters
    mu = 0
    mu_pos = 1 / np.sqrt(N)
    var = 1 / N

    recurrent_sparsity = 1 - recurrent_density
    W_rec = sparse(torch.empty(N, N), recurrent_sparsity, mu, var, generator)

    # spectral radius adjustment
    W_rec = W_rec - torch.diag(torch.diag(W_rec))
    w, v = linalg.eig(W_rec)
    spec_radius = np.max(np.absolute(w))
    W_rec = radius * W_rec / spec_radius

    W_inp = torch.zeros([N, num_inputs]).float()
    input_sparsity = 1 - input_density
    W_inp = sparse(W_inp, input_sparsity, mu_pos, var, generator)

    output_sparsity = 1 - output_density
    W_out = sparse(torch.empty(num_outputs, N), output_sparsity, mu_pos, var, generator)

    output_mask = (W_out != 0).to(device=device).float()
    input_mask = (W_inp != 0).to(device=device).float()
    recurrent_mask = torch.ones(N, N) - torch.eye(N)
    return W_rec.to(device=device).float(), \
           W_inp.to(device=device).float(), \
           W_out.to(device=device).float(), \
           recurrent_mask.to(device=device).float(), \
           output_mask.to(device=device).float(), \
           input_mask.to(device=device).float()


def get_connectivity_Dale(device, N, num_inputs, num_outputs, radius=1.5, recurrent_density=1, input_density=1,
                          output_density=1, generator=None):
    '''
    generates W_inp, W_rec and W_out matrices of RNN, with specified parameters, subject to a Dales law,
    and about 20:80 ratio of inhibitory neurons to exchitatory ones.
    Following the paper "Training Excitatory-Inhibitory Recurrent Neural Networks for Cognitive Tasks:
    A Simple and Flexible Framework" - Song et al. (2016)

    :param device: torch related: CPU or GPU
    :param N: number of neural nodes
    :param num_inputs: number of input channels, input dimension
    :param num_outputs: number of output channels, output dimension
    :param radius: spectral radius of the generated cnnectivity matrix: controls the maximal abs value of eigenvectors.
    the greater the parameter is the more sustained and chaotic activity the network exchibits, the lower - the quicker
    the network relaxes back to zero.
    :param recurrent_density: oppposite of sparcirty of the reccurrent matrix. 1.0 - fully connected recurrent matrix
    :param input_density: 1.0 - fully connected input matrix, 0 - maximally sparce matrix
    :param output_density: 1.0 - fully connected output matrix, 0 - maximally sparce matrix
    :param generator: torch random generator, for reproducibility
    :return:
    '''
    Ne = int(N * 0.8)
    Ni = int(N * 0.2)

    # Initialize W_rec
    W_rec = torch.empty([0, N])

    # Balancing parameters
    mu_E = 1 / np.sqrt(N)
    mu_I = 4 / np.sqrt(N)

    var = 1 / N
    # generating excitatory part of connectivity and an inhibitory part of connectivity:
    rowE = torch.empty([Ne, 0])
    rowI = torch.empty([Ni, 0])
    recurrent_sparsity = 1 - recurrent_density
    rowE = torch.cat((rowE, torch.abs(sparse(torch.empty(Ne, Ne), recurrent_sparsity, mu_E, var, generator))), 1)
    rowE = torch.cat((rowE, -torch.abs(sparse(torch.empty(Ne, Ni), recurrent_sparsity, mu_I, var, generator))), 1)
    rowI = torch.cat((rowI, torch.abs(sparse(torch.empty(Ni, Ne), recurrent_sparsity, mu_E, var, generator))), 1)
    rowI = torch.cat((rowI, -torch.abs(sparse(torch.empty(Ni, Ni), recurrent_sparsity, mu_I, var, generator))), 1)

    W_rec = torch.cat((W_rec, rowE), 0)
    W_rec = torch.cat((W_rec, rowI), 0)

    #  spectral radius adjustment
    W_rec = W_rec - torch.diag(torch.diag(W_rec))
    w, v = linalg.eig(W_rec)
    spec_radius = np.max(np.absolute(w))
    W_rec = radius * W_rec / spec_radius

    W_inp = torch.zeros([N, num_inputs]).float()
    input_sparsity = 1 - input_density
    W_inp = torch.abs(sparse(W_inp, input_sparsity, mu_E, var, generator))

    W_out = torch.zeros([num_outputs, N])
    output_sparsity = 1 - output_density
    W_out = torch.abs(sparse(W_out, output_sparsity, mu_E, var, generator))

    dale_mask = torch.sign(W_rec).to(device=device).float()
    output_mask = (W_out != 0).to(device=device).float()
    input_mask = (W_inp != 0).to(device=device).float()
    recurrent_mask = torch.ones(N, N) - torch.eye(N)
    return W_rec.to(device=device).float(), W_inp.to(device=device).float(), W_out.to(
        device=device).float(), recurrent_mask.to(device=device).float(), dale_mask, output_mask, input_mask


'''
Continuous-time RNN class implemented in pytorch to train with BPTT
'''


class RNN_torch(torch.nn.Module):
    def __init__(self,
                 N,
                 activation,
                 dt=1,
                 tau=10,
                 constrained=True,
                 connectivity_density_rec=1.0,
                 spectral_rad=1.2,
                 sigma_rec=.03,
                 sigma_inp=.03,
                 bias_rec=None,
                 y_init=None,
                 random_generator=None,
                 input_size=6,
                 output_size=2,
                 device=None):
        '''
        :param N: int, number of neural nodes in the RNN
        :param activation: torch function, activation function in the dynamics of the RNN
        :param constrained: whether the connectivity is constrained to comply with Dales law and elements of W_inp, W_out > 0
        :param connectivity_density_rec: float, defines the sparcity of the connectivity
        :param spectral_rad: float, spectral radius of the initial connectivity matrix W_rec
        :param dt: float, time resolution of RNN
        :param tau: float, internal time constant of the RNN-neural nodes
        :param sigma_rec: float, std of the gaussian noise in the recurrent dynamics
        :param sigma_inp: float, std of the gaussian noise in the input to the RNN
        :param bias_rec: array of N values, (inhibition/excitation of neural nodes from outside of the network)
        :param y_init: array of N values, initial value of the RNN dynamics
        :param random_generator: torch random generator, for reproducibility
        :param output_size: number of the output channels of the RNN
        :param device:
        '''
        super(RNN_torch, self).__init__()
        self.N = N
        self.activation = activation
        self.tau = tau
        self.dt = dt
        self.alpha = (dt / tau)
        self.sigma_rec = torch.tensor(sigma_rec)
        self.sigma_inp = torch.tensor(sigma_inp)
        self.input_size = input_size
        self.output_size = output_size
        self.spectral_rad = spectral_rad
        self.connectivity_density_rec = connectivity_density_rec
        self.constrained = constrained
        self.dale_mask = None

        if not (y_init is None):
            self.y_init = y_init
        else:
            self.y_init = torch.zeros(self.N)
        # self.device = torch.device('cpu')
        if (device is None):
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
            else:
                self.device = torch.device('mps') if torch.backends.mps.is_available() else torch.device('cpu')
        else:
            self.device = torch.device(device)
        print(f"Using {self.device} for RNN!")

        self.random_generator = random_generator
        self.input_layer = (torch.nn.Linear(self.input_size, self.N, bias=False)).to(self.device)
        self.recurrent_layer = torch.nn.Linear(self.N, self.N, bias=(False if (bias_rec is None) else bias_rec)).to(
            self.device)
        self.output_layer = torch.nn.Linear(self.N, self.output_size, bias=False).to(self.device)

        if self.constrained:
            # imposing a bunch of constraint on the connectivity:
            # positivity of W_inp, W_out,
            # W_rec has to be subject to Dale's law
            W_rec, W_inp, W_out, self.recurrent_mask, self.dale_mask, self.output_mask, self.input_mask = \
                get_connectivity_Dale(device, self.N, num_inputs=self.input_size, num_outputs=self.output_size,
                                      radius=self.spectral_rad, generator=self.random_generator,
                                      recurrent_density=self.connectivity_density_rec)
        else:
            W_rec, W_inp, W_out, self.recurrent_mask, self.output_mask, self.input_mask = \
                get_connectivity(device, self.N, num_inputs=self.input_size, num_outputs=self.output_size,
                                 radius=self.spectral_rad,
                                 generator=self.random_generator,
                                 recurrent_density=self.connectivity_density_rec)
        self.output_layer.weight.data = W_out.to(self.device)
        self.input_layer.weight.data = W_inp.to(self.device)
        self.recurrent_layer.weight.data = W_rec.to(self.device)
        if bias_rec is None:
            self.recurrent_layer.bias = None

    def forward(self, u, w_noise=True):
        '''
        forward dynamics of the RNN (full trial)
        :param u: array of input vectors (self.input_size, T_steps, batch_size)
        :param w_noise: bool, pass forward with or without noise
        :return: the full history of the internal variables and the outputs
        '''
        T_steps = u.shape[1]
        batch_size = u.shape[-1]
        states = torch.zeros(self.N, 1, batch_size, device=self.device)
        states[:, 0, :] = deepcopy(self.y_init).reshape(-1, 1).repeat(1, batch_size)
        rec_noise = torch.zeros(self.N, T_steps, batch_size, device=self.device)
        inp_noise = torch.zeros(self.input_size, T_steps, batch_size)
        if w_noise:
            rec_noise = torch.sqrt((2 / self.alpha) * self.sigma_rec ** 2) \
                        * torch.randn(*rec_noise.shape, generator=self.random_generator)
            inp_noise = torch.sqrt((2 / self.alpha) * self.sigma_inp ** 2) \
                        * torch.randn(*inp_noise.shape, generator=self.random_generator)
        # passing through layers require batch-first shape!
        # that's why we need to reshape the inputs and states!
        states = torch.swapaxes(states, 0, -1)
        u = torch.swapaxes(u, 0, -1).to(self.device)
        rec_noise = torch.swapaxes(rec_noise, 0, -1).to(self.device)
        inp_noise = torch.swapaxes(inp_noise, 0, -1).to(self.device)
        for i in range(T_steps - 1):
            state_new = (1 - self.alpha) * states[:, i, :] + \
                        self.alpha * (
                                self.activation(
                                    self.recurrent_layer(states[:, i, :]) +
                                    self.input_layer(u[:, i, :] + inp_noise[:, i, :])) +
                                rec_noise[:, i, :]
                        )
            states = torch.cat((states, state_new.unsqueeze_(1)), 1)
        outputs = torch.swapaxes(self.output_layer(states), 0, -1)
        states = torch.swapaxes(states, 0, -1)
        return states, outputs

    def get_params(self):
        '''
        Save crucial parameters of the RNN as numpy arrays
        :return: parameter dictionary containing connectivity parameters, initial conditions,
         number of nodes, dt and tau
        '''
        param_dict = {}
        W_out = deepcopy(self.output_layer.weight.data.cpu().detach().numpy())
        W_rec = deepcopy(self.recurrent_layer.weight.data.cpu().detach().numpy())
        W_inp = deepcopy(self.input_layer.weight.data.cpu().detach().numpy())
        y_init = deepcopy(self.y_init.detach().cpu().numpy())
        if not (self.recurrent_layer.bias is None):
            bias_rec = deepcopy(self.recurrent_layer.bias.data.cpu().detach().numpy())
        else:
            bias_rec = None
        param_dict["W_out"] = W_out
        param_dict["W_inp"] = W_inp
        param_dict["W_rec"] = W_rec
        param_dict["bias_rec"] = bias_rec
        param_dict["y_init"] = y_init
        param_dict["N"] = self.N
        param_dict["dt"] = self.dt
        param_dict["tau"] = self.tau
        return param_dict

    def set_params(self, params):
        self.output_layer.weight.data = torch.from_numpy(params["W_out"]).to(self.device)
        self.input_layer.weight.data = torch.from_numpy(params["W_inp"]).to(self.device)
        self.recurrent_layer.weight.data = torch.from_numpy(params["W_rec"]).to(self.device)
        if not (self.recurrent_layer.bias is None):
            self.recurrent_layer.bias.data = torch.from_numpy(params["bias_rec"]).to(self.device)
        self.y_init = torch.from_numpy(params["y_init"]).to(self.device)
        return None


if __name__ == '__main__':
    N = 100
    activation = lambda x: torch.maximum(x, torch.tensor(0))
    rnn_torch = RNN_torch(N=N, activation=activation, constrained=True)
    param_dict = rnn_torch.get_params()
    print(param_dict)



'''
Class which accepts RNN_torch and a task and has a mode to train RNN
'''

from copy import deepcopy

import numpy as np
import torch


def L2_ortho(rnn, X=None, y=None):
    # regularization of the input and ouput matrices
    b = torch.cat((rnn.input_layer.weight, rnn.output_layer.weight.t()), dim=1)
    b = b / torch.norm(b, dim=0)
    return torch.norm(b.t() @ b - torch.diag(torch.diag(b.t() @ b)), p=2)


def print_iteration_info(iter, train_loss, min_train_loss, val_loss, min_val_loss):
    gr_prfx = '\033[92m'
    gr_sfx = '\033[0m'

    train_prfx = gr_prfx if (train_loss <= min_train_loss) else ''
    train_sfx = gr_sfx if (train_loss <= min_train_loss) else ''
    if not (val_loss is None):
        val_prfx = gr_prfx if (val_loss <= min_val_loss) else ''
        val_sfx = gr_sfx if (val_loss <= min_val_loss) else ''
        print(f"iteration {iter},"
              f" train loss: {train_prfx}{np.round(train_loss, 6)}{train_sfx},"
              f" validation loss: {val_prfx}{np.round(val_loss, 6)}{val_sfx}")
    else:
        print(f"iteration {iter},"
              f" train loss: {train_prfx}{np.round(train_loss, 6)}{train_sfx}")


class Trainer():
    def __init__(self, RNN, Task, max_iter, tol, criterion, optimizer, lambda_orth, lambda_r):
        '''
        :param RNN: pytorch RNN (specific template class)
        :param Task: task (specific template class)
        :param max_iter: maximum number of iterations
        :param tol: float, such that if the cost function reaches tol the optimization terminates
        :param criterion: function to evaluate loss
        :param optimizer: pytorch optimizer (Adam, SGD, etc.)
        :param lambda_ort: float, regularization softly imposing a pair-wise orthogonality
         on columns of W_inp and rows of W_out
        :param lambda_r: float, regularization of the mean firing rates during the trial
        '''
        self.RNN = RNN
        self.Task = Task
        self.max_iter = max_iter
        self.tol = tol
        self.criterion = criterion
        self.optimizer = optimizer
        self.lambda_orth = lambda_orth
        self.lambda_r = lambda_r

    def train_step(self, input, target_output, mask):
        states, predicted_output = self.RNN(input)
        loss = self.criterion(target_output[:, mask, :], predicted_output[:, mask, :]) + \
               self.lambda_orth * L2_ortho(self.RNN) + \
               self.lambda_r * torch.mean(states ** 2)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        error_vect = torch.sum(((target_output[:, mask, :] - predicted_output[:, mask, :]) ** 2).squeeze(),
                               dim=1) / len(mask)
        return loss.item(), error_vect

    def eval_step(self, input, target_output, mask):
        with torch.no_grad():
            self.RNN.eval()
            states, predicted_output_val = self.RNN(input, w_noise=False)
            val_loss = self.criterion(target_output[:, mask, :], predicted_output_val[:, mask, :]) + \
                       self.lambda_orth * L2_ortho(self.RNN) + \
                       self.lambda_r * torch.mean(states ** 2)
            return float(val_loss.cpu().numpy())

    def run_training(self, train_mask, same_batch=False):
        train_losses = []
        val_losses = []
        self.RNN.train()  # puts the RNN into training mode (sets update_grad = True)
        min_train_loss = np.inf
        min_val_loss = np.inf
        best_net_params = deepcopy(self.RNN.get_params())
        if same_batch:
            input_batch, target_batch, conditions_batch = self.Task.get_batch()
            input_batch = torch.from_numpy(input_batch.astype("float32")).to(self.RNN.device)
            target_batch = torch.from_numpy(target_batch.astype("float32")).to(self.RNN.device)
            input_val = deepcopy(input_batch)
            target_output_val = deepcopy(target_batch)
            # input_val, target_output_val, conditions_val = self.Task.get_batch()
            # input_val = torch.from_numpy(input_val.astype("float32")).to(self.RNN.device)
            # target_output_val = torch.from_numpy(target_output_val.astype("float32")).to(self.RNN.device)

        for iter in range(self.max_iter):
            if not same_batch:
                input_batch, target_batch, conditions_batch = self.Task.get_batch()
                input_batch = torch.from_numpy(input_batch.astype("float32")).to(self.RNN.device)
                target_batch = torch.from_numpy(target_batch.astype("float32")).to(self.RNN.device)
                input_val, target_output_val, conditions_val = self.Task.get_batch()
                input_val = torch.from_numpy(input_val.astype("float32")).to(self.RNN.device)
                target_output_val = torch.from_numpy(target_output_val.astype("float32")).to(self.RNN.device)

            train_loss, error_vect = self.train_step(input=input_batch, target_output=target_batch, mask=train_mask)
            if self.RNN.constrained:
                # positivity of entries of W_inp and W_out
                self.RNN.output_layer.weight.data = torch.maximum(self.RNN.output_layer.weight.data, torch.tensor(0))
                self.RNN.input_layer.weight.data = torch.maximum(self.RNN.input_layer.weight.data, torch.tensor(0))
                # Dale's law
                self.RNN.recurrent_layer.weight.data = (
                            torch.maximum(self.RNN.recurrent_layer.weight.data.cpu() * self.RNN.dale_mask.cpu(),
                                          torch.tensor(0)) * self.RNN.dale_mask).to(self.RNN.device)

            # validation
            val_loss = self.eval_step(input_val, target_output_val, train_mask)
            # keeping track of train and valid losses and printing
            print_iteration_info(iter, train_loss, min_train_loss, val_loss, min_val_loss)

            train_losses.append(train_loss)
            val_losses.append(val_loss)
            if val_loss <= min_val_loss:
                min_val_loss = val_loss
                best_net_params = deepcopy(self.RNN.get_params())
            if train_loss <= min_train_loss:
                min_train_loss = train_loss

            if val_loss <= self.tol:
                self.RNN.set_params(best_net_params)
                return self.RNN, train_losses, val_losses, best_net_params

        self.RNN.set_params(best_net_params)
        return self.RNN, train_losses, val_losses, best_net_params
